"""Project Settings Service."""

from __future__ import annotations

import json
import typing as t

import structlog

from meltano.core.error import ProjectReadonly
from meltano.core.settings_service import SettingsService, SettingValueStore
from meltano.core.utils import nest_object

if t.TYPE_CHECKING:
    from meltano.core.project import Project
    from meltano.core.setting_definition import SettingDefinition

logger = structlog.get_logger(__name__)


class ProjectSettingsService(SettingsService):  # noqa: WPS214
    """Project Settings Service."""

    config_override = {}
    supports_environments = False

    def __init__(
        self,
        project: Project,
        show_hidden: bool = True,
        env_override: dict | None = None,
        config_override: dict | None = None,
    ):
        """Instantiate a `ProjectSettingsService` instance.

        Args:
            project: Meltano project instance.
            show_hidden: Whether to display secret setting values.
            env_override: Optional override environment values.
            config_override:  Optional override configuration values.
        """
        super().__init__(
            project=project,
            show_hidden=show_hidden,
            env_override=env_override,
            config_override=config_override,
        )

        # terminal env vars are already present from `SettingService.env`
        self.env_override = {
            # static, project-level env vars (e.g. MELTANO_ENVIRONMENT)
            **self.project.env,
            # env vars stored in the base `meltano.yml` `env:` key
            **self.project.meltano.env,
            # overrides
            **self.env_override,
        }

        self.config_override = {  # noqa: WPS601
            **self.__class__.config_override,
            **self.config_override,
        }

        try:
            self.ensure_project_id()
        except ProjectReadonly:
            logger.debug(
                "Cannot update `project_id` in `meltano.yml`: project is read-only.",
            )

    @property
    def project_settings_service(self):
        """Get the settings service for this project.

        For ProjectSettingsService, just returns self.

        Returns:
            self
        """
        return self

    def ensure_project_id(self) -> None:
        """Ensure `project_id` is configured properly.

        Every `meltano.yml` file should contain the `project_id`
        key-value pair. It should be present in the top-level config, rather
        than in any environment-level configs.

        If it is not present, it will be restored from `analytics.json` if possible.
        """
        try:
            project_id = self.get("project_id")
        except OSError:
            project_id = None

        if project_id is None:
            try:
                with open(
                    self.project.meltano_dir() / "analytics.json",
                ) as analytics_json_file:
                    project_id = json.load(analytics_json_file)["project_id"]
            except (OSError, KeyError, json.JSONDecodeError) as err:
                logger.debug(
                    "Unable to restore 'project_id' from 'analytics.json'",
                    err=err,
                )
            else:
                self.set("project_id", project_id, store=SettingValueStore.MELTANO_YML)
                logger.debug("Restored 'project_id' from 'analytics.json'")

    @property
    def label(self) -> str:
        """Return label.

        Returns:
            Project label.
        """
        return "Meltano"

    @property
    def docs_url(self) -> str:
        """Return docs URL.

        Returns:
            URL for Meltano doc site.
        """
        return "https://docs.meltano.com/reference/settings"

    @property
    def db_namespace(self) -> str:
        """Return namespace for setting value records in system database.

        Returns:
            Namespace for setting value records in system database.
        """
        return "meltano"

    @property
    def setting_definitions(self) -> list[SettingDefinition]:
        """Return definitions of supported settings.

        Returns:
            A list of defined settings.
        """
        return self.project.config_service.settings

    @property
    def meltano_yml_config(self):
        """Return current configuration in `meltano.yml`.

        Returns:
            Current configuration in `meltano.yml`.
        """
        return self.project.config_service.current_config

    def update_meltano_yml_config(self, config):
        """Update configuration in `meltano.yml`.

        Args:
            config: Updated config.
        """
        self.project.config_service.update_config(config)

    def process_config(self, config) -> dict:
        """Process configuration dict for presentation in `meltano config meltano`.

        Args:
            config: Config to process.

        Returns:
            Processed configuration dict for presentation in `meltano config meltano`.
        """
        return nest_object(config)



from __future__ import annotations

import json
import platform
import shutil
import typing as t

import boto3
import mock
import pytest
from moto import mock_aws

import meltano
from asserts import assert_cli_runner
from meltano.cli import cli

if t.TYPE_CHECKING:
    from click.testing import CliRunner


class TestCliUpgrade:
    @pytest.mark.usefixtures("project")
    def test_upgrade(self, cli_runner: CliRunner):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )
        # If an editable install was used, test that it cannot be upgraded automatically
        if meltano.__file__.endswith("/src/meltano/__init__.py"):
            result = cli_runner.invoke(cli, ["upgrade"])
            assert_cli_runner(result)

            assert (
                "The `meltano` package could not be upgraded automatically"
                in result.stdout
            )
            assert "run `meltano upgrade --skip-package`" in result.stdout

        with mock.patch(
            "meltano.cli.upgrade.UpgradeService._upgrade_package",
        ) as upgrade_package_mock:
            upgrade_package_mock.return_value = True

            result = cli_runner.invoke(cli, ["upgrade"])
            assert_cli_runner(result)

            assert (
                "Meltano and your Meltano project have been upgraded!" in result.stdout
            )

    @pytest.mark.usefixtures("project")
    def test_upgrade_skip_package(self, cli_runner: CliRunner):
        result = cli_runner.invoke(cli, ["upgrade", "--skip-package"])
        assert_cli_runner(result)

        assert "Your Meltano project has been upgraded!" in result.stdout

    @pytest.mark.usefixtures("project")
    def test_upgrade_package(self, cli_runner: CliRunner):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )
        # If an editable install was used, test that it cannot be upgraded automatically
        if meltano.__file__.endswith("/src/meltano/__init__.py"):
            result = cli_runner.invoke(cli, ["upgrade", "package"])
            assert_cli_runner(result)

            assert (
                "The `meltano` package could not be upgraded automatically"
                in result.stdout
            )
            assert "run `meltano upgrade --skip-package`" not in result.stdout

    @pytest.mark.order(before="test_upgrade_files_glob_path")
    @pytest.mark.usefixtures("session")
    def test_upgrade_files(self, project, cli_runner: CliRunner):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )
        result = cli_runner.invoke(cli, ["upgrade", "files"])
        output = result.stdout + result.stderr
        assert_cli_runner(result)

        assert "Nothing to update" in result.stdout

        result = cli_runner.invoke(cli, ["add", "files", "airflow"])
        output = result.stdout + result.stderr
        assert_cli_runner(result)

        # Don't update file if unchanged
        file_path = project.root_dir("orchestrate/dags/meltano.py")
        file_content = file_path.read_text()

        result = cli_runner.invoke(cli, ["upgrade", "files"])
        output = result.stdout + result.stderr
        assert_cli_runner(result)

        assert "Updating 'airflow' files in project..." in output
        assert "Nothing to update" in output
        assert file_path.read_text() == file_content

        # Update file if changed
        file_path.write_text("Overwritten!")

        # The behavior being tested assumes that the file is not locked.
        shutil.rmtree(project.root_dir("plugins/files"), ignore_errors=True)
        result = cli_runner.invoke(cli, ["upgrade", "files"])
        output = result.stdout + result.stderr
        assert_cli_runner(result)

        assert "Updated orchestrate/dags/meltano.py" in output
        assert file_path.read_text() == file_content

        # Don't update file if unchanged
        result = cli_runner.invoke(cli, ["upgrade", "files"])
        output = result.stdout + result.stderr
        assert_cli_runner(result)

        assert "Nothing to update" in output
        assert file_path.read_text() == file_content

        # Don't update file if automatic updating is
        result = cli_runner.invoke(
            cli,
            [
                "config",
                "--plugin-type",
                "files",
                "airflow",
                "set",
                "_update",
                "orchestrate/dags/meltano.py",
                "false",
            ],
        )
        output = result.stdout + result.stderr
        assert_cli_runner(result)

        file_path.write_text("Overwritten!")

        cli_runner.invoke(cli, ("lock", "airflow"))
        result = cli_runner.invoke(cli, ["upgrade", "files"])
        output = result.stdout + result.stderr
        assert_cli_runner(result)

        assert "Nothing to update" in output
        assert file_path.read_text() != file_content

        # Update file if automatic updating is re-enabled
        result = cli_runner.invoke(
            cli,
            [
                "config",
                "--plugin-type",
                "files",
                "airflow",
                "unset",
                "_update",
                "orchestrate/dags/meltano.py",
            ],
        )
        output = result.stdout + result.stderr
        assert_cli_runner(result)

        result = cli_runner.invoke(cli, ["upgrade", "files"])
        output = result.stdout + result.stderr
        assert_cli_runner(result)
        assert "Updated orchestrate/dags/meltano.py" in output

    @pytest.mark.usefixtures("session")
    def test_upgrade_files_glob_path(self, project, cli_runner: CliRunner):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )

        result = cli_runner.invoke(cli, ["add", "files", "airflow"])
        assert_cli_runner(result)

        file_path = project.root_dir("orchestrate/dags/meltano.py")
        file_path.write_text("Overwritten!")

        # override airflow--meltano.lock update extra config
        result = cli_runner.invoke(
            cli,
            [
                "config",
                "--plugin-type",
                "files",
                "airflow",
                "set",
                "_update",
                json.dumps(
                    {
                        "orchestrate/dags/meltano.py": False,
                        "*.py": True,
                    },
                ),
            ],
        )
        assert_cli_runner(result)

        result = cli_runner.invoke(cli, ["upgrade", "files"])
        output = result.stdout + result.stderr
        assert_cli_runner(result)
        assert "Updated orchestrate/dags/meltano.py" in output

    @pytest.mark.usefixtures("project")
    def test_upgrade_database(self, cli_runner: CliRunner):
        result = cli_runner.invoke(cli, ["upgrade", "database"])
        assert_cli_runner(result)

    @mock_aws
    @pytest.mark.usefixtures("project")
    def test_upgrade_state(self, cli_runner, monkeypatch):
        state_ids = [f"dev:tap-{i}-to-target-{i}" for i in range(10)]
        conn = boto3.resource("s3", region_name="us-east-1")
        bucket = conn.create_bucket(Bucket="test-state-bucket")
        for state_id in state_ids:
            bucket.put_object(
                Key=f"some/trailing/delim/path/some/trailing/delim/path/{state_id}/state.json",  # noqa: E501
            )
        monkeypatch.setenv(
            "MELTANO_STATE_BACKEND_URI",
            "s3://test-state-bucket/some/trailing/delim/path/",
        )
        result = cli_runner.invoke(cli, ["upgrade", "--skip-package"])
        assert_cli_runner(result)
        keys = [s3_object.key for s3_object in bucket.objects.all()]
        for state_id in state_ids:
            key = f"/some/trailing/delim/path/{state_id}/state.json"
            assert key in keys



"""Defines PluginTestService."""

from __future__ import annotations

import asyncio
import json
import typing as t
from abc import ABC, abstractmethod

import structlog

from meltano.core.plugin.base import PluginType
from meltano.core.plugin.error import PluginNotSupportedError

if t.TYPE_CHECKING:
    from meltano.core.plugin_invoker import PluginInvoker

logger = structlog.stdlib.get_logger(__name__)


class PluginTestServiceFactory:
    """Factory class to resolve a plugin test service."""

    def __init__(self, plugin_invoker: PluginInvoker):
        """Construct a PluginTestServiceFactory instance.

        Args:
            plugin_invoker: The invocation instance of the plugin to test.
        """
        self.plugin_invoker = plugin_invoker

    def get_test_service(self):
        """Resolve a test service instance for a plugin type.

        Returns:
            The test service instance.

        Raises:
            PluginNotSupportedError: If the plugin type is not supported for testing.
        """
        test_services = {PluginType.EXTRACTORS: ExtractorTestService}

        try:
            return test_services[self.plugin_invoker.plugin.type](self.plugin_invoker)
        except KeyError as err:
            raise PluginNotSupportedError(self.plugin_invoker.plugin) from err


class PluginTestService(ABC):
    """Abstract base class for plugin test operations."""

    def __init__(self, plugin_invoker: PluginInvoker):
        """Construct a PluginTestService instance.

        Args:
            plugin_invoker: The invocation instance of the plugin to test
        """
        self.plugin_invoker = plugin_invoker

    @abstractmethod
    async def validate(self) -> tuple[bool, str]:
        """Abstract method to validate plugin configuration."""


class ExtractorTestService(PluginTestService):
    """Handle extractor test operations."""

    async def validate(self) -> tuple[bool, str | None]:  # noqa: C901
        """Validate extractor configuration.

        Returns:
            The validation result and supporting context message (if applicable).
        """
        process = None

        try:
            process = await self.plugin_invoker.invoke_async(
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.STDOUT,
            )
        except Exception as exc:
            return False, str(exc)

        record_message_received: bool = False
        last_line: str | None = None

        while process.stdout and not process.stdout.at_eof():
            data = await process.stdout.readline()
            line = data.decode("utf-8").strip()
            if line:
                logger.debug(line)
                last_line = line

            try:
                message_type = json.loads(line)["type"]
            except (json.decoder.JSONDecodeError, KeyError):
                continue

            record_message_received = message_type in {"RECORD", "BATCH"}
            if record_message_received:
                process.terminate()
                break

        returncode = await process.wait()
        logger.debug("Process return code: %s", returncode)

        return (
            record_message_received,
            last_line if returncode else "No RECORD or BATCH message received",
        )



"""Meltano telemetry."""

from __future__ import annotations

import atexit
import json
import locale
import os
import re
import typing as t
import uuid
from contextlib import contextmanager, suppress
from datetime import datetime, timezone
from enum import Enum, auto
from urllib.parse import urlparse
from warnings import warn

import requests
import structlog
import tzlocal
from psutil import Process
from requests.adapters import HTTPAdapter
from snowplow_tracker import Emitter, SelfDescribing, SelfDescribingJson
from snowplow_tracker import Tracker as SnowplowTracker
from urllib3 import Retry

from meltano.core.tracking.schemas import (
    BlockEventSchema,
    CliEventSchema,
    ExitEventSchema,
    TelemetryStateChangeEventSchema,
)
from meltano.core.utils import format_exception

if t.TYPE_CHECKING:
    from collections.abc import Mapping
    from pathlib import Path

    from meltano.core.project import Project
    from meltano.core.tracking.contexts import (  # noqa: F401
        CliEvent,
        EnvironmentContext,
        ProjectContext,
    )

from functools import cached_property

URL_REGEX = (
    r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+"
)

MICROSECONDS_PER_SECOND = 1000000

logger = structlog.get_logger(__name__)


class BlockEvents(Enum):
    """Events describing a block state."""

    initialized = auto()
    started = auto()
    completed = auto()
    failed = auto()


def check_url(url: str) -> bool:
    """Check if the given URL is valid.

    Args:
        url: The URL to check.

    Returns:
        True if the URL is valid, False otherwise.
    """
    return bool(re.match(URL_REGEX, url))


class TelemetrySettings(t.NamedTuple):
    """Settings which control telemetry and anonymous usage stats.

    These are stored within `analytics.json`.
    """

    client_id: uuid.UUID | None
    project_id: uuid.UUID | None
    send_anonymous_usage_stats: bool | None


class Tracker:  # noqa: WPS214, WPS230 - too many (public) methods
    """Meltano tracker backed by Snowplow."""

    def __init__(  # noqa: WPS210, WPS213 - too many local variables, too many expressions
        self,
        project: Project,
        request_timeout: float | tuple[float, float] | None = 3.5,
    ):
        """Initialize a tracker for the Meltano project.

        Args:
            project: The Meltano project.
            request_timeout: Timeout for the HTTP requests. Can be set either
                as single float value which applies to both `connect` and
                `read` timeout, or as tuple with two float values which specify
                the `connect` and `read` timeouts separately.
        """
        from meltano.core.tracking.contexts import (  # noqa: WPS442, F811
            ProjectContext,
            environment_context,
        )

        self.project = project
        self.send_anonymous_usage_stats = project.settings.get(
            "send_anonymous_usage_stats",
            not project.settings.get("disable_tracking", False),
        )

        endpoints = project.settings.get("snowplow.collector_endpoints")

        session = requests.Session()

        # Retry failed requests up to 3 times, but don't retry on error status codes
        adapter = HTTPAdapter(max_retries=Retry(total=3, status=0))
        session.mount("http://", adapter)
        session.mount("https://", adapter)
        emitters: list[Emitter] = []
        for endpoint in endpoints:
            if not check_url(endpoint):
                logger.warning("invalid_snowplow_endpoint", endpoint=endpoint)
                continue
            parsed_url = urlparse(endpoint)
            emitters.append(
                Emitter(
                    endpoint=parsed_url.hostname + parsed_url.path,
                    protocol=parsed_url.scheme or "http",
                    port=parsed_url.port,
                    request_timeout=request_timeout,
                    session=session,
                ),
            )

        if emitters:
            self.snowplow_tracker = SnowplowTracker(
                namespace="meltano-core",
                app_id="meltano",
                emitters=emitters,
            )
            self.snowplow_tracker.subject.set_lang(locale.getlocale()[0])
            self.snowplow_tracker.subject.set_timezone(self.timezone_name)
            self.setup_exit_event()
        else:
            self.snowplow_tracker = None

        stored_telemetry_settings = self.load_saved_telemetry_settings()

        self.client_id = self.get_client_id(stored_telemetry_settings)

        project_ctx = ProjectContext(project, self.client_id)
        self.project_id: uuid.UUID = project_ctx.project_uuid
        self._contexts: tuple[SelfDescribingJson] = (
            environment_context,
            project_ctx,
        )

        # Environment variables to set in invoked processes for telemetry purposes
        self.env: Mapping[str, str] = {
            "MELTANO_PARENT_CONTEXT_UUID": environment_context.data["context_uuid"],
        }

        if all(setting is None for setting in stored_telemetry_settings):
            self.save_telemetry_settings()
        else:
            self.telemetry_state_change_check(stored_telemetry_settings)
            if self.client_id != stored_telemetry_settings.client_id:
                self.save_telemetry_settings()

    def get_client_id(self, stored_telemetry_settings: TelemetrySettings) -> uuid.UUID:
        """Get the telemetry client ID.

        It can be set using the `$MELTANO_CLIENT_ID` environment variable. If
        that environment variable has not been set, then the client ID stored
        on disk will be used if it exists. If the client ID has not been stored
        on disk, then a new one will be randomly generated.

        Args:
            stored_telemetry_settings: The telemetry settings stored on disk.

        Returns:
            The client ID.
        """
        with suppress(KeyError):
            uuid_str = os.environ["MELTANO_CLIENT_ID"]
            try:
                return uuid.UUID(uuid_str)
            except ValueError:
                warn(
                    (
                        f"Invalid telemetry client UUID {uuid_str!r} from "
                        "$MELTANO_CLIENT_ID"
                    ),
                    RuntimeWarning,
                    stacklevel=2,
                )
        if stored_telemetry_settings.client_id is not None:
            return stored_telemetry_settings.client_id
        return uuid.uuid4()

    @property
    def contexts(self) -> tuple[SelfDescribingJson]:
        """Get the contexts that will accompany events fired by this tracker.

        Returns:
            The contexts that will accompany events fired by this tracker.
        """
        from meltano.core.tracking.contexts import ExceptionContext

        # The `ExceptionContext` is re-created every time this is accessed
        # because it details the exceptions that are being processed when it
        # is created.
        return (*self._contexts, ExceptionContext())

    def telemetry_state_change_check(
        self,
        stored_telemetry_settings: TelemetrySettings,
    ) -> None:
        """Check prior values against current ones, and send a change event if needed.

        Args:
            stored_telemetry_settings: the prior analytics settings
        """
        if (
            stored_telemetry_settings.project_id is not None
            and stored_telemetry_settings.project_id != self.project_id
            and self.send_anonymous_usage_stats
        ):
            # Project ID has changed
            self.track_telemetry_state_change_event(
                "project_id",
                stored_telemetry_settings.project_id,
                self.project_id,
            )

        if (
            stored_telemetry_settings.send_anonymous_usage_stats is not None
            and stored_telemetry_settings.send_anonymous_usage_stats
            != self.send_anonymous_usage_stats
        ):
            # Telemetry state has changed
            self.track_telemetry_state_change_event(
                "send_anonymous_usage_stats",
                stored_telemetry_settings.send_anonymous_usage_stats,
                self.send_anonymous_usage_stats,
            )

    @cached_property
    def timezone_name(self) -> str:
        """Obtain the local timezone name.

        Examples:
            The timezone name as an IANA timezone database name:

                >>> Tracker(project).timezone_name
                'Europe/Berlin'

            The timezone name as an IANA timezone abbreviation because the full
            name was not found:

                >>> Tracker(project).timezone_name
                'CET'

        Returns:
            The local timezone as an IANA TZ database name if possible, or
            abbreviation otherwise.
        """  # noqa: F821
        try:
            return tzlocal.get_localzone_name()
        except Exception:
            return datetime.now().astimezone().tzname()

    def add_contexts(self, *extra_contexts):
        """Permanently add additional Snowplow contexts to the `Tracker`.

        Args:
            extra_contexts: The additional contexts to add to the `Tracker`.
        """
        self._contexts = (*self._contexts, *extra_contexts)

    @contextmanager
    def with_contexts(self, *extra_contexts) -> Tracker:
        """Context manager within which the `Tracker` has additional Snowplow contexts.

        Args:
            extra_contexts: The additional contexts to add to the `Tracker`.

        Yields:
            A `Tracker` with additional Snowplow contexts.
        """
        prev_contexts = self._contexts
        self._contexts = (*prev_contexts, *extra_contexts)
        try:
            yield self
        finally:
            self._contexts = prev_contexts

    def can_track(self) -> bool:
        """Check if the tracker can be used.

        Returns:
            True if the tracker can be used, False otherwise.
        """
        return self.snowplow_tracker is not None and self.send_anonymous_usage_stats

    def track_unstruct_event(self, event_json: SelfDescribingJson) -> None:
        """Fire an unstructured tracking event.

        Args:
            event_json: The SelfDescribingJson event to track. See the Snowplow
                documentation for more information.
        """
        if not self.can_track():
            return
        try:
            self.snowplow_tracker.track(
                SelfDescribing(
                    event_json=event_json,
                    context=self.contexts,
                ),
            )
        except Exception as err:
            logger.debug(
                "Failed to submit unstruct event to Snowplow, error",
                err=format_exception(err),
            )

    def track_command_event(self, event: CliEvent) -> None:
        """Fire generic command tracking event.

        Args:
            event: An member from `meltano.core.tracking.CliEvent`
        """
        self.track_unstruct_event(
            SelfDescribingJson(CliEventSchema.url, {"event": event.name}),
        )

    def track_telemetry_state_change_event(
        self,
        setting_name: str,
        from_value: uuid.UUID | str | bool | None,
        to_value: uuid.UUID | str | bool | None,
    ) -> None:
        """Fire a telemetry state change event.

        Args:
            setting_name: the name of the setting that is changing
            from_value: the old value
            to_value: the new value
        """
        from meltano.core.tracking.contexts import (  # noqa: WPS442, F811
            EnvironmentContext,
            ProjectContext,
        )

        # Save the telemetry settings to ensure this is the only telemetry
        # state change event fired for this particular setting change.
        self.save_telemetry_settings()

        if self.snowplow_tracker is None:
            # The Snowplow tracker is not available (e.g. because no endpoints are set)
            return

        logger.debug(
            (
                "Telemetry state change detected. A one-time "
                "'telemetry_state_change' event will now be sent."
            ),
            setting_name=setting_name,
        )
        if isinstance(from_value, uuid.UUID):
            from_value = str(from_value)
        if isinstance(to_value, uuid.UUID):
            to_value = str(to_value)
        event_json = SelfDescribingJson(
            TelemetryStateChangeEventSchema.url,
            {
                "setting_name": setting_name,
                "changed_from": from_value,
                "changed_to": to_value,
            },
        )
        try:
            self.snowplow_tracker.track(
                SelfDescribing(
                    event_json=event_json,
                    # If tracking is disabled, then include only the minimal
                    # Snowplow contexts required
                    context=self.contexts
                    if self.send_anonymous_usage_stats
                    else tuple(
                        ctx
                        for ctx in self.contexts
                        if isinstance(ctx, (EnvironmentContext, ProjectContext))
                    ),
                ),
            )
            self.snowplow_tracker.flush()
        except Exception as err:
            logger.debug(
                (
                    "Failed to submit 'telemetry_state_change' unstruct event "
                    "to Snowplow, error"
                ),
                err=format_exception(err),
            )

    @property
    def analytics_json_path(self) -> Path:
        """Return path to the 'analytics.json' file.

        Returns:
            Path to 'analytics.json' file.
        """
        return self.project.meltano_dir() / "analytics.json"

    def load_saved_telemetry_settings(self) -> TelemetrySettings:
        """Get settings from the 'analytics.json' file.

        Returns:
            The saved telemetry settings.
        """
        try:
            with open(self.analytics_json_path) as analytics_json_file:
                analytics = json.load(analytics_json_file)
        except (OSError, json.JSONDecodeError):
            return TelemetrySettings(None, None, None)

        if missing_keys := {
            "client_id",
            "project_id",
            "send_anonymous_usage_stats",
        } - set(analytics):
            logger.debug(
                (
                    "'analytics.json' has missing keys, and will be "
                    "overwritten with new 'analytics.json'"
                ),
                missing_keys=missing_keys,
            )
            return TelemetrySettings(None, None, None)
        return TelemetrySettings(
            self._uuid_from_str(analytics.get("client_id"), warn=True),
            self._uuid_from_str(analytics.get("project_id"), warn=True),
            analytics.get("send_anonymous_usage_stats"),
        )

    def save_telemetry_settings(self) -> None:
        """Attempt to save settings to the 'analytics.json' file."""
        try:
            with open(self.analytics_json_path, "w") as new_analytics_json_file:
                json.dump(
                    {
                        "client_id": str(self.client_id),
                        "project_id": str(self.project_id),
                        "send_anonymous_usage_stats": self.send_anonymous_usage_stats,
                    },
                    new_analytics_json_file,
                )
        except OSError as err:
            logger.debug("Unable to save 'analytics.json'", err=err)

    def _uuid_from_str(
        self,
        from_val: t.Any | None,
        warn: bool,  # noqa: WPS442
    ) -> uuid.UUID | None:
        """Safely convert string to a UUID. Return None if invalid UUID.

        Args:
            from_val: The string.
            warn: True to warn on conversion failure.

        Returns:
            A UUID object, or None if the string cannot be converted to UUID.
        """
        if not isinstance(from_val, str):
            if from_val is None:
                return None

            logger.warning(
                "Unexpected value type in 'analytics.json'",
                expected_type="string",
                value_type=type(from_val),
                value=from_val,
            )

        try:
            return uuid.UUID(from_val)
        except (ValueError, TypeError):
            # Should only be reached if user manually edits 'analytics.json'.
            log_fn = logger.warning if warn else logger.debug
            log_fn(
                "Invalid UUID string in 'analytics.json'",
                value=from_val,
            )
            return None

    def track_block_event(self, block_type: str, event: BlockEvents) -> None:
        """Fire generic block tracking event.

        Args:
            block_type: The block type.
            event: The event string (e.g. "initialize", "started", etc)
        """
        self.track_unstruct_event(
            SelfDescribingJson(
                BlockEventSchema.url,
                {"type": block_type, "event": event.name},
            ),
        )

    def setup_exit_event(self):
        """If not already done, register the atexit handler to fire the exit event.

        This method also provides this tracker instance to the CLI module for
        it to fire an exit event immediately before the CLI exits. The atexit
        handler is used only as a fallback.
        """
        from meltano import cli

        if cli.atexit_handler_registered:
            return

        cli.atexit_handler_registered = True

        # Provide `meltano.cli` with this tracker to track the exit event with
        # more context.
        cli.exit_event_tracker = self

        # As a fallback, use atexit to help ensure the exit event is sent.
        atexit.register(self.track_exit_event)

    def track_exit_event(self):
        """Fire exit event."""
        from meltano import cli

        if cli.exit_code_reported:
            return
        cli.exit_code_reported = True

        start_time = datetime.fromtimestamp(Process().create_time(), tz=timezone.utc)

        # This is the reported "end time" for this process, though in reality
        # the process will end a short time after this time as it takes time
        # to emit the event.
        now = datetime.now(timezone.utc)

        self.track_unstruct_event(
            SelfDescribingJson(
                ExitEventSchema.url,
                {
                    "exit_code": cli.exit_code,
                    "exit_timestamp": f"{now.isoformat()}",
                    "process_duration_microseconds": int(
                        (now - start_time).total_seconds() * MICROSECONDS_PER_SECOND,
                    ),
                },
            ),
        )
        self.snowplow_tracker.flush()
        atexit.unregister(self.track_exit_event)



from __future__ import annotations

import contextlib
import logging
import os
import typing as t

import pytest
import sqlalchemy as sa
from sqlalchemy import create_engine
from sqlalchemy.pool import NullPool

if t.TYPE_CHECKING:
    from sqlalchemy.engine import URL


def recreate_database(engine, db_name):
    """Drop & Create a new database.

    We need to use the master connection to do this.

    Args:
        engine: The master connection.
        db_name: The name of the database to create.
    """
    with contextlib.suppress(sa.exc.ProgrammingError), engine.begin() as conn:
        # DROP DATABASE IF EXISTS is not supported by SQL Server 2016 and up
        conn.execute(sa.text(f"DROP DATABASE IF EXISTS {db_name}"))

    with contextlib.suppress(sa.exc.ProgrammingError), engine.begin() as conn:
        conn.execute(sa.text(f"CREATE DATABASE {db_name}"))


def create_connection_url(
    host: str,
    port: int,
    user: str,
    password: str,
    database: str,
) -> URL:
    """Create a MSSQL connection URL for the given parameters.

    Args:
        host: The hostname.
        port: The port.
        user: The username.
        password: The password.
        database: The database name.

    Returns:
        The connection URL.
    """
    return sa.engine.URL.create(
        "mssql+pymssql",
        username=user,
        password=password,
        host=host,
        port=port,
        database=database,
    )


@pytest.fixture(scope="session")
def engine_uri(worker_id: str):
    host = os.getenv("MSSQL_ADDRESS")
    port = os.getenv("MSSQL_PORT", 1433)
    user = os.getenv("MSSQL_USER")
    password = os.getenv("MSSQL_PASSWORD")
    database = f"{os.getenv('MSSQL_DB', 'pytest_meltano')}_{worker_id}"

    # Recreate the database using the master database
    master_engine_uri = create_connection_url(host, port, user, password, "master")
    engine = create_engine(
        master_engine_uri,
        isolation_level="AUTOCOMMIT",
        poolclass=NullPool,
        future=True,
    )
    recreate_database(engine, database)

    # Connect to the database where the tests will be run
    testing_engine_uri = create_connection_url(host, port, user, password, database)

    return testing_engine_uri.render_as_string(hide_password=False)


@pytest.fixture()
def pg_stats(request, session):
    yield

    from meltano.core.job import Job

    jobs = session.query(Job).all()
    logging.info("%s created %d Job.", request.node.name, len(jobs))  # noqa: TID251, WPS323



from __future__ import annotations

import json
import typing as t
from collections import OrderedDict

import pytest

from meltano.core.plugin.singer.catalog import SelectedNode, SelectionType
from meltano.core.plugin.singer.tap import SingerTap
from meltano.core.select_service import SelectService

if t.TYPE_CHECKING:
    from sqlalchemy.orm.session import Session

    from meltano.core.project import Project


@pytest.mark.asyncio()
@pytest.mark.usefixtures("tap")
async def test_select_service_list_all(
    project: Project,
    session: Session,
    monkeypatch: pytest.MonkeyPatch,
):
    catalog = {
        "streams": [
            {
                "stream": "users",
                "tap_stream_id": "users",
                "metadata": [
                    {"breadcrumb": [], "metadata": {"selected": True}},
                ],
                "schema": {
                    "properties": {
                        "id": {"type": "integer"},
                    }
                },
            }
        ]
    }
    extractor = "tap-mock"
    service = SelectService(project, extractor)

    async def mock_run_discovery(tap, plugin_invoker, catalog_path):  # noqa: ARG001
        with catalog_path.open("w") as catalog_file:
            json.dump(catalog, catalog_file)

    # Mock tap's run_discovery method
    monkeypatch.setattr(
        SingerTap,
        "run_discovery",
        mock_run_discovery,
    )

    list_all = await service.list_all(session, refresh=False)
    assert list_all.streams == {
        SelectedNode(
            key="users",
            selection=SelectionType.SELECTED,
        )
    }
    assert list_all.properties == OrderedDict(
        {
            "users": {
                SelectedNode(
                    key="id",
                    selection=SelectionType.AUTOMATIC,
                ),
            }
        }
    )

    # Update the catalog to include a new property
    catalog["streams"][0]["schema"]["properties"]["name"] = {"type": "string"}  # noqa: WPS219

    # Without refreshing the catalog, the new property should not be included
    list_all = await service.list_all(session, refresh=False)
    assert list_all.properties == OrderedDict(
        {
            "users": {
                SelectedNode(
                    key="id",
                    selection=SelectionType.AUTOMATIC,
                ),
            }
        }
    )

    # Refreshing the catalog should include the new property
    list_all = await service.list_all(session, refresh=True)
    assert list_all.properties == OrderedDict(
        {
            "users": {
                SelectedNode(
                    key="id",
                    selection=SelectionType.AUTOMATIC,
                ),
                SelectedNode(
                    key="name",
                    selection=SelectionType.AUTOMATIC,
                ),
            }
        }
    )



"""Base class for all Meltano plugins."""

from __future__ import annotations

import re
from collections import defaultdict

import yaml
from structlog.stdlib import get_logger

from meltano.core.behavior import NameEq
from meltano.core.behavior.canonical import Canonical
from meltano.core.behavior.hookable import HookObject
from meltano.core.job_state import STATE_ID_COMPONENT_DELIMITER
from meltano.core.plugin.command import Command
from meltano.core.plugin.requirements import PluginRequirement
from meltano.core.setting_definition import SettingDefinition, SettingKind, YAMLEnum
from meltano.core.utils import NotFound, find_named

logger = get_logger(__name__)


class VariantNotFoundError(Exception):
    """Raised when a variant is not found."""

    def __init__(self, plugin: PluginDefinition, variant_name: str):
        """Create a new VariantNotFoundError.

        Args:
            plugin: The plugin definition.
            variant_name: The name of the variant that was not found.
        """
        self.plugin = plugin
        self.variant_name = variant_name

    def __str__(self) -> str:
        """Return a string representation of the error.

        Returns:
            The string representation of the error.
        """
        return (
            f"{self.plugin.type.descriptor.capitalize()} '{self.plugin.name}' "
            f"variant '{self.variant_name}' is not known to Meltano. "
            f"Variants: {self.plugin.variant_labels}"
        )


class PluginRefNameContainsStateIdDelimiterError(Exception):
    """A name in reference to a plugin contains the state ID component delimiter."""

    def __init__(self, name: str):
        """Create a new exception.

        Args:
            name: The name of the plugin.
        """
        super().__init__(
            f"The plugin name '{name}' cannot contain the state ID component "
            f"delimiter string '{STATE_ID_COMPONENT_DELIMITER}'",
        )


yaml.add_multi_representer(YAMLEnum, YAMLEnum.yaml_representer)


class PluginType(YAMLEnum):  # noqa: WPS214
    """The type of a plugin."""

    EXTRACTORS = "extractors"
    LOADERS = "loaders"
    TRANSFORMS = "transforms"
    ORCHESTRATORS = "orchestrators"
    TRANSFORMERS = "transformers"
    FILES = "files"
    UTILITIES = "utilities"
    MAPPERS = "mappers"
    MAPPINGS = "mappings"

    def __str__(self) -> str:
        """Return a string representation of the plugin type.

        Returns:
            The string representation of the plugin type.
        """
        return self.value

    @property
    def descriptor(self) -> str:
        """Return the descriptor of the plugin type.

        Returns:
            The descriptor of the plugin type.
        """
        return "file bundle" if self is self.__class__.FILES else self.singular

    @property
    def singular(self) -> str:
        """Make singular form for `meltano add PLUGIN_TYPE`.

        Returns:
            The singular form of the plugin type.
        """
        return "utility" if self is self.__class__.UTILITIES else self.value[:-1]

    @property
    def verb(self) -> str:
        """Make verb form.

        Returns:
            The verb form of the plugin type.
        """
        if self is self.__class__.TRANSFORMS:
            return self.singular
        if self is self.__class__.UTILITIES:
            return "utilize"
        return "map" if self is self.__class__.MAPPERS else self.value[:-3]

    @property
    def discoverable(self) -> bool:
        """Whether this plugin type is discoverable on the Hub.

        Returns:
            Whether this plugin type is discoverable on the Hub.
        """
        return self is not self.__class__.MAPPINGS

    @classmethod
    def value_exists(cls, value: str) -> bool:
        """Check if a plugin type exists.

        Args:
            value: The plugin type to check.

        Returns:
            True if the plugin type exists, False otherwise.
        """
        return value in cls._value2member_map_

    @classmethod
    def cli_arguments(cls) -> list:
        """Return the list of plugin types that can be used as CLI arguments.

        Returns:
            The list of plugin types that can be used as CLI arguments.
        """
        return [
            getattr(plugin_type, plugin_type_name)
            for plugin_type in cls
            for plugin_type_name in ("singular", "value")
        ]

    @classmethod
    def from_cli_argument(cls, value: str) -> PluginType:
        """Get the plugin type from a CLI argument.

        Args:
            value: The CLI argument.

        Returns:
            The plugin type.

        Raises:
            ValueError: If the plugin type does not exist.
        """
        for plugin_type in cls:
            if value in {plugin_type.value, plugin_type.singular}:
                return plugin_type

        raise ValueError(f"{value!r} is not a valid {cls.__name__}")  # noqa: EM102

    @classmethod
    def plurals(cls) -> list[str]:
        """Return the list of plugin plural names.

        Returns:
            The list of plugin plurals.
        """
        return [plugin_type.value for plugin_type in cls]


class PluginRef(Canonical):
    """A reference to a plugin."""

    def __init__(self, plugin_type: str | PluginType, name: str, **kwargs):
        """Create a new PluginRef.

        Args:
            plugin_type: The type of the plugin.
            name: The name of the plugin.
            kwargs: Additional keyword arguments.

        Raises:
            PluginRefNameContainsStateIdDelimiterError: If the name contains
                the state ID component delimiter string.
        """
        if STATE_ID_COMPONENT_DELIMITER in name:
            raise PluginRefNameContainsStateIdDelimiterError(name)

        self._type = (
            plugin_type
            if isinstance(plugin_type, PluginType)
            else PluginType(plugin_type)
        )

        super().__init__(name=name, **kwargs)

    @property
    def type(self) -> PluginType:
        """Return the type of the plugin.

        Returns:
            The type of the plugin.
        """
        return self._type

    def __eq__(self, other: PluginRef) -> bool:
        """Compare two plugin references.

        Args:
            other: The other plugin reference.

        Returns:
            True if the plugin references are equal, False otherwise.
        """
        return self.name == other.name and self.type == other.type

    def __hash__(self) -> int:
        """Return the hash of the plugin reference.

        Returns:
            The hash of the plugin reference.
        """
        return hash((self.type, self.name))

    def set_presentation_attrs(self, extras):
        """Set the presentation attributes of the plugin reference.

        Args:
            extras: The presentation attributes.
        """
        self.update(
            hidden=extras.pop("hidden", None),
            label=extras.pop("label", None),
            logo_url=extras.pop("logo_url", None),
            description=extras.pop("description", None),
        )


class Variant(NameEq, Canonical):
    """A variant of a plugin."""

    ORIGINAL_NAME = "original"
    DEFAULT_NAME = "default"

    def __init__(
        self,
        name: str | None = None,
        original: bool | None = None,
        deprecated: bool | None = None,
        docs: str | None = None,
        repo: str | None = None,
        pip_url: str | None = None,
        python: str | None = None,
        executable: str | None = None,
        description: str | None = None,
        logo_url: str | None = None,
        capabilities: list | None = None,
        settings_group_validation: list | None = None,
        settings: list | None = None,
        commands: dict | None = None,
        requires: dict[PluginType, list] | None = None,
        env: dict[str, str] | None = None,
        **extras,
    ):
        """Create a new Variant.

        Args:
            name: The name of the variant.
            original: Whether the variant is the original one.
            deprecated: Whether the variant is deprecated.
            docs: The documentation URL.
            repo: The repository URL.
            pip_url: The pip URL.
            python: The path to the Python executable to use, or name to find on the
                $PATH. Defaults to the Python executable running Meltano.
            executable: The executable name.
            description: The description of the plugin.
            logo_url: The logo URL of the plugin.
            capabilities: The capabilities of the variant.
            settings_group_validation: The settings group validation.
            settings: The settings of the variant.
            commands: The commands of the variant.
            requires: Other plugins this plugin depends on.
            env: Environment variables to inject into plugins runtime context.
            extras: Additional keyword arguments.
        """
        super().__init__(
            name=name,
            original=original,
            deprecated=deprecated,
            docs=docs,
            repo=repo,
            pip_url=pip_url,
            python=python,
            executable=executable,
            description=description,
            logo_url=logo_url,
            capabilities=list(capabilities or []),
            settings_group_validation=list(settings_group_validation or []),
            settings=list(map(SettingDefinition.parse, settings or [])),
            commands=Command.parse_all(commands),
            requires=PluginRequirement.parse_all(requires),
            env=env or {},
            extras=extras,
        )


class PluginDefinition(PluginRef):
    """A plugin definition."""

    def __init__(
        self,
        plugin_type: PluginType,
        name: str,
        namespace: str,
        *,
        python: str | None = None,
        variant: str | None = None,
        variants: list | None = None,
        is_default_variant: bool | None = None,
        **extras,
    ):
        """Create a new PluginDefinition.

        Args:
            plugin_type: The type of the plugin.
            name: The name of the plugin.
            namespace: The namespace of the plugin.
            python: The path to the Python executable to use, or name to find on the
                $PATH. Defaults to the Python executable running Meltano.
            variant: The variant of the plugin.
            variants: The variants of the plugin.
            is_default_variant: Whether the variant is the default one.
            extras: Additional keyword arguments.
        """
        super().__init__(plugin_type, name)

        self._defaults["label"] = lambda plugin: plugin.name

        def default_logo_url(plugin_def):
            short_name = re.sub(
                r"^(tap|target)-",  # noqa: WPS360
                "",
                plugin_def.name,
            )
            return f"/static/logos/{short_name}-logo.png"

        self._defaults["logo_url"] = default_logo_url

        if not variants:
            variant = Variant(variant, **extras)

            # Any properties considered "extra" by the variant should be
            # considered extras of the plugin definition.
            extras = variant.extras
            variant.extras = {}

            variants = [variant]

        # Attributes will be listed in meltano.yml in this order:
        self.namespace = namespace
        self.python = python
        self.set_presentation_attrs(extras)
        self.extras = extras
        self.variants = [Variant.parse(x) for x in variants]
        self.is_default_variant = is_default_variant

    def __iter__(self):
        """Iterate over the variants of the plugin definition.

        Yields:
            The variants of the plugin definition.
        """
        for attr, value in super().__iter__():
            if attr == "variants" and len(value) == 1:
                # If there is only a single variant, its properties can be
                # nested in the plugin definition
                for variant_k, variant_v in value[0]:
                    if variant_k == "name":
                        variant_k = "variant"

                    yield (variant_k, variant_v)
            else:
                yield (attr, value)

    def get_variant(self, variant_name: str) -> Variant:
        """Get the variant with the given name.

        Args:
            variant_name: The name of the variant.

        Returns:
            The variant with the given name.

        Raises:
            VariantNotFoundError: If the variant is not found.
        """
        try:
            return find_named(self.variants, variant_name)
        except NotFound as err:
            raise VariantNotFoundError(self, variant_name) from err

    def find_variant(self, variant_or_name: str | Variant | None = None):
        """Find the variant with the given name or variant.

        Args:
            variant_or_name: The variant or name of the variant.

        Returns:
            The variant with the given name or variant.
        """
        if isinstance(variant_or_name, Variant):
            return variant_or_name

        if variant_or_name is None or variant_or_name == Variant.DEFAULT_NAME:
            return self.variants[0]

        if variant_or_name == Variant.ORIGINAL_NAME:
            try:
                return next(variant for variant in self.variants if variant.original)
            except StopIteration:
                return self.variants[0]

        return self.get_variant(variant_or_name)

    def variant_label(self, variant):
        """Return label for specified variant.

        Args:
            variant: The variant.

        Returns:
            The label for the variant.
        """
        variant = self.find_variant(variant)

        label = variant.name or Variant.ORIGINAL_NAME

        if variant == self.variants[0] and self.is_default_variant is not False:
            return f"{label} (default)"

        if variant.deprecated:
            return f"{label} (deprecated)"

        return label

    @property
    def variant_labels(self):
        """Return labels for supported variants.

        Returns:
            The labels for the supported variants.
        """
        return ", ".join([self.variant_label(variant) for variant in self.variants])

    @classmethod
    def from_standalone(
        cls: type[PluginDefinition],
        plugin: StandalonePlugin,
    ) -> PluginDefinition:
        """Create a new PluginDefinition from a StandalonePlugin.

        Args:
            plugin: The plugin.

        Returns:
            The new PluginDefinition.
        """
        return cls(
            plugin.plugin_type,
            plugin.name,
            plugin.namespace,
            variant=plugin.variant,
            # Extras
            label=plugin.label,
            docs=plugin.docs,
            repo=plugin.repo,
            pip_url=plugin.pip_url,
            python=plugin.python,
            executable=plugin.executable,
            description=plugin.description,
            logo_url=plugin.logo_url,
            capabilities=plugin.capabilities,
            settings_group_validation=plugin.settings_group_validation,
            settings=plugin.settings,
            commands=plugin.commands,
            requires=plugin.requires,
            env=plugin.env,
            **plugin.extras,
        )


class BasePlugin(HookObject):  # noqa: WPS214
    """A base plugin."""

    EXTRA_SETTINGS = []

    def __init__(self, plugin_def: PluginDefinition, variant: Variant):
        """Create a new BasePlugin.

        Args:
            plugin_def: The plugin definition.
            variant: The variant.
        """
        super().__init__()

        self._plugin_def = plugin_def
        self._variant = variant

    def __eq__(self, other: BasePlugin):
        """Compare two plugins.

        Args:
            other: The other plugin.

        Returns:
            True if the plugins are equal, False otherwise.
        """
        return (
            self._plugin_def == other._plugin_def  # noqa: WPS437
            and self._variant == other._variant  # noqa: WPS437
        )

    def __hash__(self) -> int:
        """Return the hash of the plugin.

        Returns:
            The hash of the plugin.
        """
        return hash((self._plugin_def, self._variant))

    def __iter__(self):
        """Iterate over the settings of the plugin.

        Yields:
            The settings of the plugin.
        """
        yield from self._plugin_def

    def __getattr__(self, attr: str):
        """Get the value of the setting.

        Args:
            attr: The name of the setting.

        Returns:
            The value of the setting.
        """
        try:
            return getattr(self._plugin_def, attr)
        except AttributeError:
            return getattr(self._variant, attr)

    @property
    def variant(self) -> str:
        """Return the variant name.

        Returns:
            The variant name.
        """
        return self._variant.name

    @property
    def executable(self) -> str:
        """Return the plugin executable.

        Returns:
            The executable name.
        """
        return self._variant.executable or self._plugin_def.name

    @property
    def extras(self):
        """Return the plugin extras.

        Returns:
            The extras.
        """
        return {**self._plugin_def.extras, **self._variant.extras}

    @property
    def all_commands(self):
        """Return a dictionary of supported commands.

        Returns:
            A dictionary of supported commands.
        """
        return self._variant.commands

    @property
    def test_commands(self) -> dict[str, Command]:
        """Return a the test command for this plugin.

        Returns:
            The test command for this plugin.
        """
        return {
            name: command
            for name, command in self.all_commands.items()
            if name.startswith("test")
        }

    @property
    def all_settings(self):
        """Return a list of settings.

        Returns:
            A list of settings.
        """
        return self._variant.settings

    @property  # noqa: WPS210
    def extra_settings(self):  # noqa: WPS210
        """Return the extra settings for this plugin.

        Returns:
            The extra settings for this plugin.
        """
        defaults = {f"_{name}": value for name, value in self.extras.items()}

        existing_settings = []
        for setting in self.EXTRA_SETTINGS:
            default_value = defaults.get(setting.name)
            if default_value is not None:
                setting = setting.with_attrs(value=default_value)

            existing_settings.append(setting)

        # Create setting definitions for unknown defaults,
        # including flattened keys of default nested object items
        existing_settings.extend(
            SettingDefinition.from_missing(
                existing_settings,
                defaults,
                custom=False,
                default=True,
            ),
        )

        return existing_settings

    @property
    def all_requires(self):
        """Return a list of requires.

        Returns:
            A list of requires.
        """
        return self._variant.requires

    def env_prefixes(
        self,
        for_writing=False,  # noqa: ARG002
    ) -> list[str]:
        """Return environment variable prefixes to use for settings.

        Args:
            for_writing: Whether to return environment variable prefixes for writing.

        Returns:
            The environment variable prefixes.
        """
        return [self.name, self.namespace]

    def is_installable(self) -> bool:
        """Return whether the plugin is installable.

        Returns:
            True if the plugin is installable, False otherwise.
        """
        return self.pip_url is not None

    def is_invokable(self) -> bool:
        """Return whether the plugin is invokable.

        Returns:
            Whether the plugin is invokable.
        """
        return self.is_installable() or self.executable is not None

    def is_configurable(self) -> bool:
        """Return whether the plugin is configurable.

        Returns:
            True if the plugin is configurable, False otherwise.
        """
        return True

    def should_add_to_file(self) -> bool:
        """Return whether the plugin should be added to the config file.

        Returns:
            True if the plugin should be added to the config file, False otherwise.
        """
        return True

    def exec_args(
        self,
        files: dict,  # noqa: ARG002
    ):
        """Return the arguments to pass to the plugin runner.

        Args:
            files: The files to pass to the plugin runner.

        Returns:
            The arguments to pass to the plugin runner.
        """
        return []

    @property
    def config_files(self):
        """Return a list of stubbed files created for this plugin.

        Returns:
            A list of stubbed files created for this plugin.
        """
        return {}

    @property
    def output_files(self):
        """Return a list of stubbed files created for this plugin.

        Returns:
            A list of stubbed files created for this plugin.
        """
        return {}

    def process_config(self, config):
        """Process the config for this plugin.

        Args:
            config: The config to process.

        Returns:
            The processed config.
        """
        return config

    @property
    def definition(self) -> PluginDefinition:
        """Return the plugin definition.

        Returns:
            The plugin definition.
        """
        return self._plugin_def


class StandalonePlugin(Canonical):
    """A standalone plugin definition representing a single variant."""

    def __init__(
        self,
        plugin_type: str,
        name: str,
        namespace: str,
        variant: str | None = None,
        label: str | None = None,
        docs: str | None = None,
        repo: str | None = None,
        pip_url: str | None = None,
        python: str | None = None,
        executable: str | None = None,
        description: str | None = None,
        logo_url: str | None = None,
        capabilities: list | None = None,
        settings_group_validation: list | None = None,
        settings: list | None = None,
        commands: dict | None = None,
        requires: dict[PluginType, list] | None = None,
        env: dict[str, str] | None = None,
        **extras,
    ):
        """Create a locked plugin.

        Args:
            plugin_type: The plugin type.
            name: The name of the plugin.
            namespace: The namespace of the plugin.
            variant: The variant of the plugin.
            label: The label of the plugin.
            docs: The documentation URL of the plugin.
            repo: The repository URL of the plugin.
            pip_url: The pip URL of the plugin.
            python: The path to the Python executable to use, or name to find on the
                $PATH. Defaults to the Python executable running Meltano.
            executable: The executable of the plugin.
            description: The description of the plugin.
            logo_url: The logo URL of the plugin.
            capabilities: The capabilities of the plugin.
            settings_group_validation: The settings group validation of the plugin.
            settings: The settings of the plugin.
            commands: The commands of the plugin.
            requires: Other plugins this plugin depends on.
            env: Environment variables to inject into plugins runtime context.
            extras: Additional attributes to set on the plugin.
        """
        super().__init__(
            plugin_type=PluginType(plugin_type),
            name=name,
            namespace=namespace,
            variant=variant,
            label=label,
            docs=docs,
            repo=repo,
            pip_url=pip_url,
            python=python,
            executable=executable,
            description=description,
            logo_url=logo_url,
            capabilities=capabilities or [],
            settings_group_validation=settings_group_validation or [],
            settings=list(map(SettingDefinition.parse, settings or [])),
            commands=Command.parse_all(commands),
            requires=PluginRequirement.parse_all(requires),
            env=env or {},
            extras=extras,
        )

        deprecated_kind_replacements = {
            SettingKind.HIDDEN: "hidden: true",
            SettingKind.PASSWORD: "sensitive: true",
        }

        settings_by_deprecated_kind = defaultdict(list)

        for s in self.settings:
            if s.kind in deprecated_kind_replacements:
                settings_by_deprecated_kind[s.kind].append(s)

        if settings_by_deprecated_kind:
            for kind, settings in settings_by_deprecated_kind.items():
                logger.warning(
                    f"`kind: {kind}` is deprecated for setting definitions"  # noqa: G003
                    + (
                        f" in favour of `{deprecated_kind_replacements[kind]}`"
                        if deprecated_kind_replacements[kind]
                        else ""
                    )
                    + ", and is currently in use by the following settings of "
                    + f"{self.plugin_type.singular} '{self.name}': "
                    + ", ".join([f"'{s.name}'" for s in settings])
                    + ". "
                    + "Please open an issue or pull request to update the plugin "
                    + "definition on Meltano Hub at "
                    + f"https://github.com/meltano/hub/blob/main/_data/meltano/{self.plugin_type}/{self.name}/{self.variant}.yml.",  # noqa: E501
                )

    @classmethod
    def from_variant(
        cls: type[StandalonePlugin],
        variant: Variant,
        plugin_def: PluginDefinition,
    ):
        """Create a locked plugin from a variant and plugin definition.

        Args:
            variant: The variant to create the plugin from.
            plugin_def: The plugin definition to create the plugin from.

        Returns:
            A locked plugin definition.
        """
        return cls(
            plugin_type=plugin_def.type,
            name=plugin_def.name,
            namespace=plugin_def.namespace,
            variant=variant.name,
            label=plugin_def.label,
            docs=variant.docs,
            repo=variant.repo,
            pip_url=variant.pip_url,
            python=plugin_def.python,
            executable=variant.executable,
            description=variant.description,
            logo_url=variant.logo_url,
            capabilities=variant.capabilities,
            settings_group_validation=variant.settings_group_validation,
            settings=variant.settings,
            commands=variant.commands,
            requires=variant.requires,
            env=variant.env,
            **{**plugin_def.extras, **variant.extras},
        )



"""StateStoreManager for Azure Blob storage backend."""

from __future__ import annotations

import typing as t
from contextlib import contextmanager
from functools import cached_property

from meltano.core.error import MeltanoError
from meltano.core.state_store.filesystem import (
    CloudStateStoreManager,
)

if t.TYPE_CHECKING:
    from collections.abc import Iterator

AZURE_INSTALLED = True

try:
    from azure.storage.blob import BlobServiceClient
except ImportError:
    AZURE_INSTALLED = False


class MissingAzureError(Exception):
    """Raised when azure is required but no installed."""

    def __init__(self):
        """Initialize a MissingAzureError."""
        super().__init__(
            "azure required but not installed. Install meltano[azure] to use Azure Blob Storage as a state backend.",  # noqa: E501
        )


@contextmanager
def requires_azure():
    """Raise MissingAzureError if azure is required but missing in context.

    Raises:
        MissingAzureError: if azure is not installed.

    Yields:
        None
    """
    if not AZURE_INSTALLED:
        raise MissingAzureError
    yield


class AZStorageStateStoreManager(CloudStateStoreManager):
    """State backend for Azure Blob Storage."""

    label: str = "Azure Blob Storage"

    def __init__(
        self,
        connection_string: str | None = None,
        prefix: str | None = None,
        storage_account_url: str | None = None,
        **kwargs,
    ):
        """Initialize the BaseFilesystemStateStoreManager.

        Args:
            connection_string: connection string to use in authenticating to Azure
            prefix: the prefix to store state at
            storage_account_url: url of the azure stroga account
            kwargs: additional keyword args to pass to parent

        Raises:
            MeltanoError: If container name is not included in the URI.
        """
        super().__init__(**kwargs)
        self.connection_string = connection_string
        self.storage_account_url = storage_account_url

        if not self.parsed.hostname:
            raise MeltanoError(
                f"Azure state backend URI must include a container name: {self.uri}",  # noqa: EM102
                "Verify state backend URI. Must be in the form of azure://<container>/<prefix>",  # noqa: E501
            )

        self.container_name = self.parsed.hostname
        self.prefix = prefix or self.parsed.path

    @staticmethod
    def is_file_not_found_error(err: Exception) -> bool:
        """Check if err is equivalent to file not being found.

        Args:
            err: the err to check

        Returns:
            True if error represents file not being found, else False
        """
        from azure.core.exceptions import ResourceNotFoundError

        return (
            isinstance(err, ResourceNotFoundError)
            and "ErrorCode:BlobNotFound" in err.args[0]
        )

    @cached_property
    def client(self) -> BlobServiceClient:
        """Get an authenticated azure.storage.blob.BlobServiceClient.

        Returns:
            An authenticated azure.storage.blob.BlobServiceClient

        Raises:
            MeltanoError: If connection string is not provided.
        """
        with requires_azure():
            if self.storage_account_url:
                from azure.identity import DefaultAzureCredential

                default_credential = DefaultAzureCredential()
                return BlobServiceClient(
                    self.storage_account_url,
                    credential=default_credential,
                )

            if self.connection_string:
                return BlobServiceClient.from_connection_string(self.connection_string)

            raise MeltanoError(
                "Azure state backend requires a connection string "  # noqa: EM101
                "or an account URL to use host credentials",
                "Read https://learn.microsoft.com/en-us/azure/storage/common/storage-configure-connection-string for more information.",  # noqa: E501
            )

    def delete(self, file_path: str):
        """Delete the file/blob at the given path.

        Args:
            file_path: the path to delete.

        Raises:
            Exception: if error not indicating file is not found is thrown
        """
        blob_client = self.client.get_blob_client(
            container=self.container_name,
            blob=file_path,
        )
        try:
            blob_client.delete_blob()
        except Exception as e:
            if not self.is_file_not_found_error(e):
                raise e

    def list_all_files(self) -> Iterator[str]:
        """List all files in the backend.

        Yields:
            The next file in the backend.
        """
        container_client = self.client.get_container_client(self.container_name)
        for blob in container_client.list_blobs(  # noqa: WPS526
            name_starts_with=self.prefix.lstrip("/"),
        ):
            yield blob.name

    def copy_file(self, src: str, dst: str) -> None:
        """Copy a file from one location to another.

        Args:
            src: the source path
            dst: the destination path
        """
        container_client = self.client.get_container_client(self.container_name)
        src_blob_client = container_client.get_blob_client(src)
        dst_blob_client = container_client.get_blob_client(dst)
        dst_blob_client.start_copy_from_url(src_blob_client.url)



"""Schema creation CLI."""

from __future__ import annotations

import click

from meltano.cli.params import pass_project
from meltano.cli.utils import InstrumentedCmd, InstrumentedGroup
from meltano.core.db import ensure_schema_exists, project_engine


@click.group(cls=InstrumentedGroup, short_help="Manage system DB schema.")
def schema():
    """Manage system DB schema."""


@schema.command(
    cls=InstrumentedCmd,
    short_help="Create system DB schema, if not exists.",
)
@click.argument("schema")
@click.argument("roles", nargs=-1, required=True)
@pass_project()
def create(project, schema_name, roles):
    """Create system DB schema, if not exists."""
    engine, _ = project_engine(project)
    ensure_schema_exists(engine, schema_name, grant_roles=roles)



from __future__ import annotations

import io
from textwrap import dedent

import pytest
from ruamel.yaml.comments import CommentedMap

from meltano.core.behavior.canonical import Canonical
from meltano.core.yaml import yaml

definition = {
    # {'a': None, 'b': 1, 'c': None, 'd': 3, 'e': None, ...}
    chr(ord("a") + i): i if i % 2 else None
    for i in range(10)
}


class TestCanonical:
    @pytest.fixture()
    def subject(self):
        return Canonical(**definition)

    def test_canonical(self, subject):
        # make sure the Nones are removed
        assert len(list(subject)) == 5

        subject.test = "hello"
        yaml_definition = "\n".join(f"{k}: {v}" for k, v in iter(subject))

        buf = io.StringIO()
        yaml.dump(subject, buf)
        buf.seek(0)

        assert buf.read().strip() == yaml_definition

    def test_false(self, subject):
        subject.false_value = False

        assert subject.canonical()["false_value"] is False

    def test_nested(self, subject):
        nested = Canonical(test="value")
        subject.nested = nested

        assert Canonical.as_canonical(subject)["nested"] == Canonical.as_canonical(
            nested,
        )

    def test_nested_empty(self, subject):
        nested = Canonical(test="")
        subject.nested = nested

        assert "nested" not in Canonical.as_canonical(subject)

    def test_update_canonical(self, subject):
        subject.update(Canonical(test="value"))
        assert subject.test == "value"

    def test_update_dict(self, subject):
        subject.update({"test": "value"})
        assert subject.test == "value"

    def test_update_kwargs(self, subject):
        subject.update(test="value")
        assert subject.test == "value"

    def test_with_attrs(self, subject):
        subject.test = "value"

        assert subject.with_attrs().canonical() == subject.canonical()

        new = subject.with_attrs(test="other_value")
        assert new.test == "other_value"
        assert new.canonical() == {**subject.canonical(), "test": "other_value"}

        new = subject.with_attrs(new_test="new_value")
        assert new.new_test == "new_value"
        assert new.canonical() == {**subject.canonical(), "new_test": "new_value"}

    def test_defaults(self, subject):
        with pytest.raises(AttributeError):
            subject.test  # noqa: B018, WPS428

        subject.test = None

        assert subject.test is None

        # This would typically be set from a Canonical subclass
        subject._defaults["test"] = lambda _: "default"

        # Default values show up when getting an attr
        assert subject.test == "default"
        # But they're not included in the canonical representation
        assert "test" not in subject.canonical()

        subject.test = "changed"

        assert subject.test == "changed"
        assert subject.canonical()["test"] == "changed"

    def test_fallbacks(self, subject):
        # Calling an unknown attribute is not supported
        with pytest.raises(AttributeError):
            subject.unknown  # noqa: B018, WPS428

        fallback = Canonical(unknown="value", known="value")
        # This would typically be set from a Canonical subclass
        subject._fallback_to = fallback

        # Unknown attributes fall back
        assert subject.unknown == "value"
        assert "unknown" not in subject.canonical()

        # Known attributes don't fall back
        subject.known = None
        assert subject.known is None

        # Unless we make them
        subject._fallbacks.add("known")
        assert subject.known == "value"
        assert "known" not in subject.canonical()

        # Unless there is nothing to fallback to
        subject._fallback_to = None
        assert subject.known is None

        # Defaults are still applied
        subject._defaults["known"] = lambda _: "default"
        assert subject.known == "default"
        assert "known" not in subject.canonical()

        # Until a value is set
        subject.known = "value"
        assert subject.known == "value"
        assert subject.canonical()["known"] == "value"

    def test_preserve_comments(self):
        contents = """\
            # This is a top-level comment
            test: value

            object:
              # Comment in an object
              key: value # Comment in a nested value

            array:
            # Comment in an array
            - value # Comment in an array element
        """
        contents = dedent(contents)
        mapping = yaml.load(io.StringIO(contents))
        subject = Canonical.parse(mapping)
        assert subject.test == "value"
        assert subject.object["key"] == "value"
        assert subject.array[0] == "value"

        obj = subject.canonical()
        assert isinstance(obj, CommentedMap)

        out_stream = io.StringIO()
        yaml.dump(obj, out_stream)
        out_stream.seek(0)
        new_contents = out_stream.read()
        assert new_contents == contents

    def test_annotations(self):
        original = CommentedMap(
            {"a": 1, "annotations": {"cloud": {"data": 123}}, "z": -1},
        )
        obj = Canonical.parse(original)
        assert obj.a == 1
        with pytest.raises(AttributeError):
            assert obj.annotations
        assert obj.z == -1
        assert obj.canonical() == original



from __future__ import annotations

import typing as t

import pytest

from meltano.core.job import Job, Payload
from meltano.core.plugin import PluginType
from meltano.core.project_plugins_service import PluginAlreadyAddedException

if t.TYPE_CHECKING:
    from meltano.core.project_add_service import ProjectAddService


class TestSingerTarget:
    @pytest.fixture()
    def subject(self, project_add_service: ProjectAddService):
        try:
            return project_add_service.add(PluginType.LOADERS, "target-mock")
        except PluginAlreadyAddedException as err:
            return err.plugin

    @pytest.mark.asyncio()
    async def test_exec_args(self, subject, session, plugin_invoker_factory):
        invoker = plugin_invoker_factory(subject)
        async with invoker.prepared(session):
            assert subject.exec_args(invoker) == ["--config", invoker.files["config"]]

    @pytest.mark.asyncio()
    async def test_setup_bookmark_writer(
        self,
        subject,
        session,
        plugin_invoker_factory,
        elt_context_builder,
    ):
        job = Job(job_name="pytest_test_runner")

        # test noop run outside of pipeline context
        invoker = plugin_invoker_factory(subject, context=None)
        async with invoker.prepared(session):
            subject.setup_bookmark_writer(invoker)
            assert invoker.output_handlers is None

        # test noop run with job but no session
        elt_context = (
            elt_context_builder.with_session(None)
            .with_loader(subject.name)
            .with_job(job)
            .context()
        )

        invoker = plugin_invoker_factory(subject, context=elt_context)
        async with invoker.prepared(session):
            subject.setup_bookmark_writer(invoker)
            assert invoker.output_handlers is None

        # test noop run with session but no job
        elt_context = (
            elt_context_builder.with_session(session)
            .with_loader(subject.name)
            .with_job(None)
            .context()
        )

        invoker = plugin_invoker_factory(subject, context=elt_context)
        async with invoker.prepared(session):
            subject.setup_bookmark_writer(invoker)
            assert invoker.output_handlers is None

        # test running inside of pipeline context
        elt_context = (
            elt_context_builder.with_session(session)
            .with_loader(subject.name)
            .with_job(job)
            .context()
        )

        invoker = plugin_invoker_factory(subject, context=elt_context)
        async with invoker.prepared(session):
            subject.setup_bookmark_writer(invoker)
            assert len(invoker.output_handlers.get(invoker.StdioSource.STDOUT)) == 1
            assert (
                invoker.output_handlers.get(invoker.StdioSource.STDOUT)[0].payload_flag
                is Payload.STATE
            )

        # test w/ incomplete state (requires a filter and full refresh)
        elt_context = (
            elt_context_builder.with_session(session)
            .with_loader(subject.name)
            .with_job(job)
            .with_select_filter(["entity", "!other_entity"])
            .with_full_refresh(True)
            .context()
        )

        invoker = plugin_invoker_factory(subject, context=elt_context)
        async with invoker.prepared(session):
            subject.setup_bookmark_writer(invoker)
            assert len(invoker.output_handlers.get(invoker.StdioSource.STDOUT)) == 1
            assert (
                invoker.output_handlers.get(invoker.StdioSource.STDOUT)[0].payload_flag
                is Payload.INCOMPLETE_STATE
            )



"""Defines helpers for the core codebase."""

from __future__ import annotations

import asyncio
import collections
import functools
import hashlib
import math
import os
import platform
import re
import sys
import traceback
import typing as t
import unicodedata
from contextlib import suppress
from copy import copy, deepcopy
from datetime import date, datetime, time, timezone
from enum import IntEnum
from functools import reduce
from operator import setitem
from pathlib import Path

import flatten_dict
import structlog
from requests.auth import HTTPBasicAuth

from meltano.core.error import MeltanoError

logger = structlog.stdlib.get_logger(__name__)

TRUTHY = ("true", "1", "yes", "on")
REGEX_EMAIL = r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)"


try:
    # asyncio.Task.all_tasks() is fully moved to asyncio.all_tasks() starting
    # with Python 3.9. Also applies to current_task.
    asyncio_all_tasks = asyncio.all_tasks
except AttributeError:
    asyncio_all_tasks = asyncio.Task.all_tasks


class NotFound(Exception):
    """An element is not found."""

    def __init__(self, name, obj_type=None):
        """Create a new exception.

        Args:
            name: the name of the element that is not found
            obj_type: the type of element
        """
        if obj_type is None:
            super().__init__(f"{name} was not found.")
        else:
            super().__init__(f"{obj_type.__name__} '{name}' was not found.")


def run_async(func: t.Callable[..., t.Coroutine[t.Any, t.Any, t.Any]]):
    """Run the given async function using `asyncio.run`.

    Args:
        func: The function to run asynchronously.

    Returns:
        The given function wrapped so as to run within `asyncio.run`.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return asyncio.run(func(*args, **kwargs))

    return wrapper


# from https://github.com/jonathanj/compose/blob/master/compose.py
def compose(*fs: t.Callable[[t.Any], t.Any]):
    """Create a composition of unary functions.

    Args:
        fs: Unary functions to compose.

    Examples:
        ```python
        compose(f, g)(x) == f(g(x))
        ```

    Returns:
        The composition of the provided unary functions, which itself is a
        unary function.
    """
    return functools.reduce(lambda f, g: lambda x: f(g(x)), compact(fs), lambda x: x)


# from http://www.dolphmathews.com/2012/09/slugify-string-in-python.html
def slugify(s):
    """Normalize strings into something URL-friendly.

    Args:
        s: the string to slugify

    Returns:
        The string as a slug

    >>> slugify("[Some] _ Article's Title--")
    'some-articles-title'
    """
    # "[Some] _ Article's Title--" -> "[some] _ article's title--"
    s = s.lower()

    # "[some] _ article's_title--" -> "[some]___article's_title__"
    for c in " -./":
        s = s.replace(c, "_")

    # "[some]___article's_title__" -> "some___articles_title__"
    s = re.sub(r"\W", "", s)

    # "some___articles_title__" -> "some   articles title  "
    s = s.replace("_", " ")

    # "some   articles title  " -> "some articles title "
    s = re.sub(r"\s+", " ", s)

    # "some articles title " -> "some articles title"
    s = s.strip()

    # "some articles title" -> "some-articles-title"
    return s.replace(" ", "-")


def get_basic_auth(user, token):
    return HTTPBasicAuth(user, token)


def pop_all(keys, d: dict):
    return {k: d.pop(k) for k in keys}


def get_all(keys, d: dict, default=None):
    return {k: d.get(k, default) for k in keys}


# Taken from https://stackoverflow.com/a/20666342
def merge(src, dest):
    """Merge both given dictionaries together at depth, modifying `dest` in-place.

    Args:
        src: A dictionary to merge into `dest`.
        dest: The dictionary that will be updated with the keys and values from
            `src` at depth.

    Examples:
        >>> a = {'f' :{'all_rows': {'pass': 'dog', 'n': '1'}}}
        >>> b = {'f' :{'all_rows': {'fail': 'cat', 'n': '5'}}}
        >>> merge(b, a) == {'f': {'all_rows': {'pass': 'dog', 'fail': 'cat', 'n': '5'}}}
        True

    Returns:
        The `dest` dictionary with the keys and values from `src` merged in.
    """
    for key, value in src.items():
        if isinstance(value, dict):
            # get node or create one
            node = dest.setdefault(key, {})
            merge(value, node)
        else:
            dest[key] = value

    return dest


def are_similar_types(left, right):
    return isinstance(left, type(right)) or isinstance(right, type(left))


def nest(d: dict, path: str, value=None, maxsplit=-1, force=False):  # noqa: WPS210
    """Create a hierarchical dictionary path and return the leaf dict.

    Args:
        d: the dictionary to operate on
        path: the dot-delimited path to operate on
        value: the value to set at the given path
        maxsplit: maximum number of splits to split path by
        force: if true, write an empty dict

    Returns:
        The leaf element of the dict

    Examples:
        >>> d = dict()
        >>> test = nest(d, "foo.bar.test")
        >>> test
        {}
        >>> d
        {'foo': {'bar': {'test': {}}}}
        >>> test["a"] = 1
        >>> d
        {'foo': {'bar': {'test': {'a': 1}}}}
        >>> alist = nest(d, "foo.bar.list", value=[])
        >>> alist.append("works")
        >>> d
        {'foo': {'bar': {'test': {'a': 1}}, 'list': ["works"]}}
    """  # noqa: P102
    if value is None:
        value = {}

    if isinstance(path, str):
        path = path.split(".", maxsplit=maxsplit)

    *initial, tail = path

    # create the list of dicts
    cursor = d
    for key in initial:
        if key not in cursor or (not isinstance(cursor[key], dict) and force):
            cursor[key] = {}

        cursor = cursor[key]

    if tail not in cursor or (
        (not are_similar_types(cursor[tail], value)) and force  # noqa: WPS516
    ):
        # We need to copy the value to make sure
        # the `value` parameter is not mutated.
        cursor[tail] = deepcopy(value)

    return cursor[tail]


def nest_object(flat_object):
    obj = {}
    for key, value in flat_object.items():
        nest(obj, key, value)
    return obj


def to_env_var(*xs: str) -> str:
    """Convert a list of strings to an environment variable name.

    Args:
        *xs: the strings to convert

    Returns:
        The environment variable name.

    Examples:
        >>> to_env_var("foo", "bar")
        'FOO_BAR'
        >>> to_env_var("foo", "bar", "baz")
        'FOO_BAR_BAZ'
        >>> to_env_var("foo.bar")
        'FOO_BAR'
    """
    return "_".join(re.sub("[^A-Za-z0-9]", "_", x).upper() for x in xs if x)


def flatten(d: dict, reducer: str | t.Callable = "tuple", **kwargs):
    """Flatten a dictionary with `dot` and `env_var` reducers.

    Wrapper around `flatten_dict.flatten`.

    Args:
        d: the dict to flatten
        reducer: the reducer to flatten with
        **kwargs: additional kwargs to pass to flatten_dict.flatten

    Returns:
        the flattened dict
    """
    if reducer == "dot":
        reducer = lambda *xs: xs[1] if xs[0] is None else ".".join(xs)  # noqa: E731
    if reducer == "env_var":
        reducer = to_env_var

    return flatten_dict.flatten(d, reducer, **kwargs)


def compact(xs: t.Iterable) -> t.Iterable:
    """Remove None values from an iterable.

    Args:
        xs: the iterable to operate on

    Returns:
        The iterable with Nones removed
    """
    return (x for x in xs if x is not None)


def file_has_data(file: Path | str):
    file = Path(file)  # ensure it is a Path object
    return file.exists() and file.stat().st_size > 0


def identity(x):
    return x


def noop(*_args, **_kwargs):
    pass


async def async_noop(*_args, **_kwargs) -> bool:
    return True


def truthy(val: str) -> bool:
    return str(val).lower() in TRUTHY


@t.overload
def coerce_datetime(d: None) -> None: ...  # noqa: WPS428


@t.overload
def coerce_datetime(d: datetime) -> datetime: ...  # noqa: WPS428


@t.overload
def coerce_datetime(d: date) -> datetime: ...  # noqa: WPS428


def coerce_datetime(d):
    """Add a `time` component to `d` if it is missing.

    Args:
        d: the date or datetime to add the time to

    Returns:
        The resulting datetime
    """
    if d is None:
        return None

    return d if isinstance(d, datetime) else datetime.combine(d, time())


@t.overload
def iso8601_datetime(d: None) -> None: ...  # noqa: WPS428


@t.overload
def iso8601_datetime(d: str) -> datetime: ...  # noqa: WPS428


def iso8601_datetime(d: str | None):
    if d is None:
        return None

    isoformats = [
        "%Y-%m-%dT%H:%M:%SZ",  # noqa: WPS323
        "%Y-%m-%dT%H:%M:%S+00:00",  # noqa: WPS323
        "%Y-%m-%dT%H:%M:%S",  # noqa: WPS323
        "%Y-%m-%d %H:%M:%S",  # noqa: WPS323
        "%Y-%m-%d",  # noqa: WPS323
    ]

    for format_string in isoformats:
        with suppress(ValueError):
            return coerce_datetime(
                datetime.strptime(d, format_string).replace(tzinfo=timezone.utc),
            )
    raise ValueError(f"{d} is not a valid UTC date.")  # noqa: EM102


class _GetItemProtocol(t.Protocol):
    def __getitem__(self, key: str) -> str: ...  # noqa: WPS428


_G = t.TypeVar("_G", bound=_GetItemProtocol)


def find_named(xs: t.Iterable[_G], name: str, obj_type: type | None = None) -> _G:
    """Find an object by its 'name' key.

    Args:
        xs: Some iterable of objects against which that name should be matched.
        name: Used to match against the input objects.
        obj_type: Object type used for generating the exception message.

    Returns:
        The first item matched, if any. Otherwise raises an exception.

    Raises:
        NotFound: If an object with the given name was not found.
    """
    try:
        return next(x for x in xs if x["name"] == name)
    except StopIteration as stop:
        raise NotFound(name, obj_type) from stop


def makedirs(func):
    @functools.wraps(func)
    def decorate(*args, **kwargs):
        enabled = kwargs.pop("make_dirs", True)

        path = func(*args, **kwargs, make_dirs=enabled)

        if not enabled:
            return path

        # if there is an extension, only create the base dir
        _, ext = os.path.splitext(path)
        os.makedirs(os.path.dirname(path) if ext else path, exist_ok=True)
        return path

    return decorate


def is_email_valid(value: str):
    return re.match(REGEX_EMAIL, value)


def pop_at_path(d, path, default=None):  # noqa: WPS210
    if isinstance(path, str):
        path = path.split(".")

    *initial, tail = path

    cursor = d
    cursors = []
    for key in initial:
        cursors.append((cursor, key))

        try:
            cursor = cursor[key]
        except KeyError:
            return default

    popped = cursor.pop(tail, default)

    for cursor, key in reversed(cursors):
        if len(cursor[key]) == 0:
            cursor.pop(key, None)

    return popped


def set_at_path(d, path, value):
    if isinstance(path, str):
        path = path.split(".")

    *initial, tail = path

    final = nest(d, initial, force=True) if initial else d
    final[tail] = value


class EnvironmentVariableNotSetError(MeltanoError):
    """A referenced environment variable is not set."""

    def __init__(self, env_var: str):
        """Initialize the error.

        Args:
            env_var: The unset environment variable name.
        """
        self.env_var = env_var

        reason = f"Environment variable '{env_var}' referenced but not set"
        instruction = "Make sure the environment variable is set"
        super().__init__(reason, instruction)


ENV_VAR_PATTERN = re.compile(
    r"""
    (?P<escape>\\)?  # escape
    \$  # starts with a '$'
    (?:
        {(?P<curly>\w+)} # ${VAR}
        |
        (?P<normal>[A-Z][A-Z0-9_]*) # $VAR
    )
    """,
    re.VERBOSE,
)

Expandable = t.TypeVar(
    "Expandable",
    str,
    t.Mapping[str, "Expandable"],
    t.Sequence["Expandable"],
)


class EnvVarMissingBehavior(IntEnum):
    """The behavior that should be employed when expanding a missing env var."""

    use_empty_str = 0
    raise_exception = 1
    ignore = 2


def expand_env_vars(
    raw_value: Expandable,
    env: t.Mapping[str, str],
    *,
    if_missing: EnvVarMissingBehavior = EnvVarMissingBehavior.use_empty_str,
    flat: bool = False,
) -> Expandable:
    """Expand/interpolate provided env vars into a string or env mapping.

    By default, attempting to expand an env var which is not defined in the
    provided env dict will result in it being replaced with the empty string.

    Args:
        raw_value: A string or env mapping in which env vars will be expanded.
        env: The env vars to use for the expansion of `raw_value`.
        if_missing: The behavior to employ if an env var in `raw_value` is not
            set in `env`.
        flat: Whether the `raw_value` has a flat structure. Ignored if
            `raw_value` is not a mapping. Otherwise it controls whether this
            function will process nested levels within `raw_value`. Defaults to
            `False` for backwards-compatibility. Setting to `True` is recommend
            for performance, safety, and cleanliness reasons.

    Raises:
        EnvironmentVariableNotSetError: Attempted to expand an env var that was not
            defined in the provided env dict, and `if_missing` was
            `EnvVarMissingBehavior.raise_exception`.

    Returns:
        The string or env dict with env vars expanded. For backwards
        compatibility, if anything other than an `str` or mapping is provided
        as the `raw_value`, it is returned unchanged.
    """  # noqa: DAR402
    if_missing = EnvVarMissingBehavior(if_missing)

    if not isinstance(raw_value, (str, t.Mapping, list)):
        return raw_value

    def replacer(match: re.Match) -> str:
        # The variable can be in either group
        var = match["curly"] or match["normal"]
        restored = f"${{{var}}}" if match["curly"] else f"${var}"
        if match["escape"]:
            return restored
        try:
            val = str(env[var])
        except KeyError as ex:
            logger.debug(
                f"Variable '${var}' is not set in the provided env dictionary.",  # noqa: G004
            )
            if if_missing == EnvVarMissingBehavior.raise_exception:
                raise EnvironmentVariableNotSetError(var) from ex
            if if_missing == EnvVarMissingBehavior.ignore:
                return restored
            return ""
        if not val:
            logger.debug(f"Variable '${var}' is empty.")  # noqa: G004
        return val

    return _expand_env_vars(raw_value, replacer, flat)


# Separate inner-function for `expand_env_vars` for performance reasons. Like
# this the `replacer` function closure only needs to be created once when
# `raw_value` is a dict, as opposed to once per key-value pair.
def _expand_env_vars(
    raw_value: Expandable,
    replacer: t.Callable[[re.Match], str],
    flat: bool,
) -> Expandable:
    if isinstance(raw_value, t.Mapping):
        if flat:
            return {k: ENV_VAR_PATTERN.sub(replacer, v) for k, v in raw_value.items()}
        return {
            k: _expand_env_vars(v, replacer, flat)
            if isinstance(v, (str, t.Mapping, list))
            else v
            for k, v in raw_value.items()
        }
    if isinstance(raw_value, list):
        # `flat=True` doesn't seem to be used anywhere and probably doesn't make sense
        # for lists anyway, so we don't support it here.
        return [
            _expand_env_vars(v, replacer, flat)
            if isinstance(v, (str, t.Mapping, list))
            else v
            for v in raw_value
        ]
    return ENV_VAR_PATTERN.sub(replacer, raw_value)


T = t.TypeVar("T")


def uniques_in(original: t.Sequence[T]) -> list[T]:
    """Get unique elements from an iterable while preserving order.

    Args:
        original: A sequence from which only the unique values will be returned.

    Returns:
        A list of unique values from the provided sequence in the order they appeared.
    """
    return list(collections.OrderedDict.fromkeys(original))


# https://gist.github.com/cbwar/d2dfbc19b140bd599daccbe0fe925597#gistcomment-2845059
def human_size(num, suffix="B"):
    """Return human-readable file size.

    Args:
        num: the number to convert
        suffix: the suffix to append to the resulting file size

    Returns:
        File size in human-readable format
    """
    magnitude = int(math.floor(math.log(num, 1024)))
    val = num / math.pow(1024, magnitude)

    if magnitude == 0:
        return f"{val:.0f} bytes"
    if magnitude > 7:
        return f"{val:.1f}Yi{suffix}"

    prefix = ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"][magnitude]
    return f"{val:3.1f}{prefix}{suffix}"


def hash_sha256(value: str | bytes) -> str:
    """Get the sha256 hash of a string.

    Args:
        value: the string value to hash.

    Returns:
        The hashed value of the given string.

    Raises:
        ValueError: If we are blindly passed a value that is None.
    """
    if value is None:
        raise ValueError("Cannot hash None.")  # noqa: EM101
    if isinstance(value, str):
        value = value.encode()
    return hashlib.sha256(value).hexdigest()


def format_exception(exception: BaseException) -> str:
    """Get the exception with its traceback formatted as it would have been printed.

    Args:
        exception: The exception value to be turned into a string.

    Returns:
        A string that shows the exception object as it would have been printed
        had it been raised and not caught.
    """
    return "".join(
        traceback.format_exception(type(exception), exception, exception.__traceback__),
    )


def safe_hasattr(obj: t.Any, name: str) -> bool:
    """Safely checks if an object has a given attribute.

    This is a hacky workaround for the fact that `hasattr` is not allowed by WPS.

    Args:
        obj: The object to check.
        name: The name of the attribute to check.

    Returns:
        True if the object has the attribute, False otherwise.
    """
    try:
        getattr(obj, name)
    except AttributeError:
        return False
    return True


def strtobool(val: str) -> bool:
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.

    Case is ignored in string comparisons.

    Re-implemented from distutils.util.strtobool to avoid importing distutils.

    Args:
        val: The string to convert to a boolean.

    Returns:
        True if the string represents a truthy value, False otherwise.

    Raises:
        ValueError: If the string is not a valid representation of a boolean.
    """
    val = val.lower()
    if val in {"y", "yes", "t", "true", "on", "1"}:
        return True
    if val in {"n", "no", "f", "false", "off", "0"}:
        return False

    raise ValueError(f"invalid truth value {val!r}")  # noqa: EM102


def get_boolean_env_var(env_var: str, default: bool = False) -> bool:
    """Get the value of an environment variable as a boolean.

    Args:
        env_var: The name of the environment variable.
        default: The default value to return if the environment variable is not set.

    Returns:
        The value of the environment variable as a boolean.
    """
    try:
        return strtobool(os.getenv(env_var, str(default)))
    except ValueError:
        return default


def get_no_color_flag() -> bool:
    """Get the truth value of the `NO_COLOR` environment variable.

    Returns:
        Whether the `NO_COLOR` environment variable is set to a truthy value.
    """
    return get_boolean_env_var("NO_COLOR")


class MergeStrategy(t.NamedTuple):
    """Strategy to be used when merging instances of a type.

    The first value of this tuple, `applicable_for_instance_of`, is a type or
    tuple of types for which the behavior (see below) should apply. An
    `isinstance` check is performed on each yet-unprocessed value using these
    types.

    The second value of this tuple, `behavior`, is a function which is provided
    the dictionary being merged into, the key into that dictionary that is
    being affected, the value for which
    `isinstance(value, applicable_for_instance_of)` was true, and the tuple of
    merge strategies in use (provided to enable recursion by calling
    `deep_merge`).

    If the behavior function returns `NotImplemented` then it will be skipped,
    and later items in the tuple of merge strategies will be tried instead.
    """

    applicable_for_instance_of: type | tuple[type, ...]
    behavior: t.Callable[
        [t.MutableMapping[str, t.Any], str, t.Any, tuple[MergeStrategy, ...] | None],
        None,
    ]


@t.runtime_checkable
class Extendable(t.Protocol):
    """A type protocol for types which have an `extend` method."""

    def extend(self, x: t.Any) -> None:
        """Extend the current instance with another value.

        Args:
            x: A value to extend this instance with.
        """


default_deep_merge_strategies: tuple[MergeStrategy, ...] = (
    MergeStrategy(
        t.Mapping,
        lambda x, k, v, s: setitem(
            x,
            k,
            _deep_merge(x.setdefault(k, v.__class__()), v, strategies=s),
        ),
    ),
    MergeStrategy(
        Extendable,
        lambda x, k, v, _: x.setdefault(k, v.__class__()).extend(v),
    ),
    MergeStrategy(object, lambda x, k, v, _: setitem(x, k, v)),
)


TMapping = t.TypeVar("TMapping", bound=t.Mapping)


def deep_merge(
    *data: TMapping,
    strategies: tuple[MergeStrategy, ...] = default_deep_merge_strategies,
) -> TMapping:
    """Merge multiple mappings at depth.

    Args:
        data: The mappings.
        strategies: A tuple of merge strategies, which are pairs of
            `(applicable type, behavior function)`. Each type will be tried in
            order until one passes an `isinstance` check for the value being
            merged, then the associated behavior function will be called to
            perform the merge. Refer to the documentation for `MergeStrategy`
            for more details. By default, the merge strategies will merge
            mappings with a recursive deep merge, objects with an `extend`
            method (e.g. `lists`) using the `extend` method, and all other
            objects with `setitem(dict_being_merged_into, key, value)`.

    Returns:
        The merged mapping.
    """
    return reduce(lambda a, b: _deep_merge(a, b, strategies), data)


def _deep_merge(a, b, strategies):
    base: TMapping = copy(a)
    for key, value in b.items():
        for applicable_types, behavior in strategies:
            if (
                isinstance(value, applicable_types)
                and behavior(base, key, value, strategies) is not NotImplemented
            ):
                break
    return base


def remove_suffix(string: str, suffix: str) -> str:
    """Remove suffix from string.

    Compatible with Python 3.8

    Args:
        string: the string to remove suffix from
        suffix: the suffix to remove

    Returns:
        The changed string
    """
    if sys.version_info >= (3, 9):
        return string.removesuffix(suffix)
    if string.endswith(suffix):
        return string[: -len(suffix)]
    return string


_filename_restriction_pattern = re.compile(r"[^\w.-]")
_reserved_windows_filenames = frozenset(
    (
        "AUX",
        "COM1",
        "COM2",
        "COM3",
        "COM4",
        "CON",
        "LPT1",
        "LPT2",
        "LPT3",
        "NUL",
        "PRN",
    ),
)
_sanitize_filename_transformations = (
    # Normalize unicode data in the string:
    lambda x: unicodedata.normalize("NFKD", x),
    # Limit the string to ASCII characters:
    lambda x: x.encode("ascii", "ignore").decode("ascii"),
    # Replace each path separator with a space:
    lambda x: x.replace(os.path.sep, " "),
    lambda x: x.replace(os.path.altsep, " ") if os.path.altsep else x,
    # Replace each whitespace character with an underscore:
    lambda x: "_".join(x.split()),
    # Limit the string to alphanumeric characters, underscores, hyphens, and dots:
    lambda x: _filename_restriction_pattern.sub("", x),
    # Remove Remove illegal character combination `._` from front and back:
    lambda x: x.strip("._"),
    # Add a leading `_` if necessary to avoid conflict with reserved Windows filenames:
    lambda x: f"_{x}"
    if platform.system() == "Windows"
    and x
    and x.split(".")[0].upper() in _reserved_windows_filenames
    else x,
)


def sanitize_filename(filename: str) -> str:
    """Sanitize `filename` in a consistent cross-platform way.

    Args:
        filename: The name of the file to sanitize - not the full path.

    Returns:
        The provided filename after a series of sanitization steps. It will
        only contain ASCII characters. If necessary on Windows, the filename
        will be prefixed by an underscore to avoid conflict with reserved
        Windows file names.
    """
    return functools.reduce(
        lambda x, y: y(x),  # noqa: WPS442
        _sanitize_filename_transformations,
        filename,
    )



"""Defines helpers related to the system database."""

from __future__ import annotations

import time
import typing as t
from urllib.parse import urlparse

import structlog
from sqlalchemy import create_engine
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import NullPool
from sqlalchemy.sql import text

from meltano.core.error import MeltanoError

if t.TYPE_CHECKING:
    from sqlalchemy.engine import Connection, Engine

    from meltano.core.project import Project

# Keep a Project → Engine mapping to serve
# the same engine for the same Project
_engines = {}
logger = structlog.stdlib.get_logger(__name__)


class MeltanoDatabaseCompatibilityError(MeltanoError):
    """Raised when the database is not compatible with Meltano."""

    INSTRUCTION = (
        "Upgrade your database to be compatible with Meltano or use a different "
        "database"
    )

    def __init__(self, reason: str):
        """Initialize the error with a reason.

        Args:
            reason: The reason why the database is not compatible.
        """
        super().__init__(reason, self.INSTRUCTION)


class NullConnectionStringError(MeltanoError):
    """Raised when the database is not compatible with Meltano."""

    REASON = "The `database_uri` setting has a null value"
    INSTRUCTION = (
        "Verify that the `database_uri` setting points to a valid database connection "
        "URI, or use `MELTANO_FF_STRICT_ENV_VAR_MODE=1 meltano config meltano list` "
        "to check for missing environment variables"
    )

    def __init__(self):
        """Initialize the exception."""
        super().__init__(self.REASON, self.INSTRUCTION)


def project_engine(
    project: Project,
    default: bool = False,
) -> tuple[Engine, sessionmaker]:
    """Create and register a SQLAlchemy engine for a Meltano project instance.

    Args:
        project: The Meltano project that the engine will be connected to.
        default: Whether the engine created should be stored as the default
            engine for this project.

    Returns:
        The engine, and a session maker bound to the engine.

    Raises:
        NullConnectionStringError: The `database_uri` setting has a null value.
    """
    if existing_engine := _engines.get(project):
        return existing_engine

    database_uri = project.settings.get("database_uri")
    parsed_db_uri = urlparse(database_uri)
    sanitized_db_uri = parsed_db_uri._replace(  # noqa: WPS437
        netloc=(
            f"{parsed_db_uri.username}:********@"  # user:pass auth case
            if parsed_db_uri.password
            else "********@"  # token auth case
            if parsed_db_uri.username
            else ""  # no auth case
        )
        + (parsed_db_uri.hostname or ""),
    ).geturl()
    logger.debug(
        f"Creating DB engine for project at {str(project.root)!r} "  # noqa: G004
        f"with DB URI {sanitized_db_uri!r}",
    )

    if database_uri is None:
        raise NullConnectionStringError

    engine = create_engine(database_uri, poolclass=NullPool, future=True)

    # Connect to the database to ensure it is available.
    connect(
        engine,
        max_retries=project.settings.get("database_max_retries"),
        retry_timeout=project.settings.get("database_retry_timeout"),
    )

    check_database_compatibility(engine)
    init_hook(engine)

    engine_session = (engine, sessionmaker(bind=engine, future=True))

    if default:
        # register the default engine
        _engines[project] = engine_session

    return engine_session


def connect(
    engine: Engine,
    max_retries: int,
    retry_timeout: float,
) -> Connection:
    """Connect to the database.

    Args:
        engine: The DB engine with which the check will be performed.
        max_retries: The maximum number of retries that will be attempted.
        retry_timeout: The number of seconds to wait between retries.

    Raises:
        OperationalError: Error during DB connection - max retries exceeded.

    Returns:
        A connection to the database.
    """
    attempt = 0
    while True:
        try:
            return engine.connect()
        except OperationalError:
            if attempt >= max_retries:
                logger.error(
                    f"Could not connect to the database after {attempt} "  # noqa: G004
                    "attempts. Max retries exceeded.",
                )
                raise
            attempt += 1
            logger.info(
                f"DB connection failed. Will retry after {retry_timeout}s. "  # noqa: G004
                f"Attempt {attempt}/{max_retries}",
            )
            time.sleep(retry_timeout)


init_hooks: dict[str, t.Callable[[Connection], None]] = {
    "sqlite": lambda x: x.execute(text("PRAGMA journal_mode=WAL")),
}


def init_hook(engine: Engine) -> None:
    """Run the initialization hook for the provided DB engine.

    The initialization hooks are taken from the `meltano.core.db.init_hooks`
    dictionary, which maps the dialect name of the engine to a unary function
    which will be called with the provided DB engine.

    Args:
        engine: The engine for which the init hook will be run.

    Raises:
        Exception: The init hook raised an exception.
    """
    if hook := init_hooks.get(engine.dialect.name):
        with engine.connect() as conn:
            try:
                hook(conn)
            except Exception as ex:
                raise Exception(f"Failed to initialize database: {ex!s}") from ex  # noqa: EM102


def ensure_schema_exists(
    engine: Engine,
    schema_name: str,
    grant_roles: tuple[str] = (),
) -> None:
    """Ensure the specified `schema_name` exists in the database.

    Args:
        engine: The DB engine to be used.
        schema_name: The name of the schema.
        grant_roles: Roles to grant to the specified schema.
    """
    group_identifiers = ",".join(grant_roles)

    create_schema = text(f"CREATE SCHEMA IF NOT EXISTS {schema_name}")
    grant_select_schema = text(
        f"ALTER DEFAULT PRIVILEGES IN SCHEMA {schema_name} GRANT SELECT ON "
        f"TABLES TO {group_identifiers}",
    )
    grant_usage_schema = text(
        f"GRANT USAGE ON SCHEMA {schema_name} TO {group_identifiers}",
    )

    with engine.connect() as conn, conn.begin():
        conn.execute(create_schema)
        if grant_roles:
            conn.execute(grant_select_schema)
            conn.execute(grant_usage_schema)

    logger.info(f"Schema {schema_name} has been created successfully.")  # noqa: G004
    for role in grant_roles:
        logger.info(f"Usage has been granted for role: {role}.")  # noqa: G004


def check_database_compatibility(engine: Engine) -> None:
    """Check that the database is compatible with Meltano.

    Args:
        engine: The DB engine to be used. This should already be connected to
            the database.

    Raises:
        MeltanoDatabaseCompatibilityError: The database is not compatible with
            Meltano.
    """
    dialect = engine.dialect.name
    version = engine.dialect.server_version_info

    if dialect == "sqlite" and version < (3, 25, 1):
        version_string = ".".join(map(str, version))
        reason = (
            f"Detected SQLite {version_string}, but Meltano requires at least 3.25.1"
        )
        raise MeltanoDatabaseCompatibilityError(reason)



"""Various small utilities for working with futures."""

from __future__ import annotations

import asyncio
from asyncio import Task

import structlog

from meltano.core.runner import RunnerError
from meltano.core.utils import human_size

logger = structlog.stdlib.get_logger(__name__)


def all_done(tasks: list[Task], done: set[Task]) -> bool:
    """Iterate through a task list checking if ALL tasks are in the done set."""
    return all(idx in done for idx in tasks)


def first_failed_future(exception_future: Task, done: set[Task]) -> Task | None:
    """Check if a future is completed and return the first failed (if any).

    Args:
        exception_future: The future you want to check and who's futures should
            be returned if an exception was raised.
        done: The set of completed futures you want to search.

    Returns:
        The first future that failed.
    """
    if exception_future in done:
        futures_done, _ = exception_future.result()
        if futures_failed := [
            future for future in futures_done if future.exception() is not None
        ]:
            return futures_failed.pop()
    return None


def handle_producer_line_length_limit_error(
    exception: Exception,
    line_length_limit: int,
    stream_buffer_size: int,
):
    """Handle `asyncio.LimitOverrunError` from producers.

    Args:
        exception: The exception to handle, which should be a `ValueError` with
            a `asyncio.LimitOverrunError` as its context.
        line_length_limit: The message size limit.
        stream_buffer_size: The stream buffer size.

    Raises:
        RunnerError: An exception raised from the `asyncio.LimitOverrunError`.
    """
    # TODO: reuse from runner/singer.py
    # StreamReader.readline can raise a ValueError wrapping a LimitOverrunError:
    # https://github.com/python/cpython/blob/v3.8.7/Lib/asyncio/streams.py#L549
    if not isinstance(exception, ValueError):
        return

    exception = exception.__context__  # noqa: WPS609
    if not isinstance(exception, asyncio.LimitOverrunError):
        return

    logger.error(
        "The extractor generated a message exceeding the message size limit "  # noqa: G004
        f"of {human_size(line_length_limit)} (half the buffer size "
        f"of {human_size(stream_buffer_size)}).",
    )
    logger.error(
        "To let this message be processed, increase the 'elt.buffer_size' "
        "setting to at least double the size of the largest expected message, "
        "and try again.",
    )
    logger.error(
        "To learn more, visit "
        "https://docs.meltano.com/reference/settings#eltbuffer_size",
    )
    raise RunnerError("Output line length limit exceeded") from exception  # noqa: EM101



"""Output Logger."""

from __future__ import annotations

import asyncio
import logging
import os
import sys
from contextlib import (
    asynccontextmanager,
    contextmanager,
    redirect_stderr,
    redirect_stdout,
    suppress,
)

import structlog

from meltano.core.runner import RunnerError

from .formatters import LEVELED_TIMESTAMPED_PRE_CHAIN
from .utils import capture_subprocess_output


class OutputLogger:
    """Output Logger."""

    def __init__(self, file):
        """Instantiate an Output Logger.

        Args:
            file: A file to output to.
        """
        self.file = file
        self.stdout = sys.stdout
        self.stderr = sys.stderr

        self.outs = {}

    def out(
        self,
        name: str,
        logger=None,
        write_level: int = logging.INFO,
    ) -> Out:
        """Obtain an Out instance for use as a logger or use for output capture.

        Args:
            name: name of this Out instance and to use in the name field.
            logger: logger to temporarily add a handler too.
            write_level: log level passed to underlying logger.log calls.

        Returns:
            An Out instance that will log anything written in to a
            file by injecting a log handler, as well as to the provided logger.
        """
        if not logger:
            logger = structlog.stdlib.get_logger()

        out = Out(
            self,
            name,
            logger=logger,
            write_level=write_level,
            file=self.file,
        )
        self.outs[name] = out
        return out


class LineWriter:
    """Line Writer."""

    def __init__(self, out):
        """Instantiate a Line Writer.

        Args:
            out: The location to write to.
        """
        self.__out = out

    def __getattr__(self, name):
        """Get attribute.

        Args:
            name: The attribute name.

        Returns:
            The specified attributes value.
        """
        return getattr(self.__out, name)

    def write(self, line):
        """Write a line.

        Args:
            line: A line to write.
        """
        self.__out.writeline(line.rstrip())


class FileDescriptorWriter:
    """File Descriptor Writer."""

    def __init__(self, out, fd):
        """Instantiate File Descriptor Writer.

        Args:
            out: Output location.
            fd: A file descriptor.
        """
        self.__out = out
        self.__writer = os.fdopen(fd, "w")

    def __getattr__(self, name):
        """Get attribute.

        Args:
            name: The attribute name.

        Returns:
            The specified attributes value.
        """
        return getattr(self.__writer, name)

    def isatty(self):
        """Is out location a tty.

        Returns:
            True if output location is a tty.
        """
        return self.__out.isatty()


class Out:  # noqa: WPS230
    """Simple Out class to log anything written in a stream."""

    def __init__(
        self,
        output_logger: OutputLogger,
        name: str,
        logger: structlog.stdlib.BoundLogger,
        write_level: int,
        file: str,
    ):
        """Log anything written in a stream.

        Args:
            output_logger: the OutputLogger to use.
            name: name of this Out instance and to use in the name field.
            logger: logger to temporarily add a handler too.
            write_level: log level passed to logger.log calls.
            file: file to associate with the FileHandler to log to.
        """
        self.output_logger = output_logger
        self.logger = logger
        self.name = name
        self.write_level = write_level
        self.file = file

        self.last_line = ""

    @property
    def redirect_log_handler(self) -> logging.Handler:
        """Configure a logging.Handler suitable for redirecting logs too.

        Returns:
            logging.FileHandler using an uncolorized console formatter
        """
        formatter = structlog.stdlib.ProcessorFormatter(
            processor=structlog.dev.ConsoleRenderer(
                colors=False,
                exception_formatter=structlog.dev.plain_traceback,
            ),
            foreign_pre_chain=LEVELED_TIMESTAMPED_PRE_CHAIN,
        )
        handler = logging.FileHandler(self.file, delay=True)
        handler.setFormatter(formatter)
        return handler

    @contextmanager
    def line_writer(self):
        """Yield a line writer instance.

        Yields:
            A line writer instance
        """
        yield LineWriter(self)

    @contextmanager
    def redirect_logging(self, ignore_errors=()):
        """Redirect log entries to a temporarily added file handler.

        Args:
            ignore_errors: A tuple of Error classes to ignore.

        Yields:
            With the side-effect of redirecting logging.
        """  # noqa: DAR401
        logger = logging.getLogger()  # noqa: TID251
        logger.addHandler(self.redirect_log_handler)
        ignored_errors = (
            KeyboardInterrupt,
            asyncio.CancelledError,
            *ignore_errors,
        )
        try:
            yield
        except ignored_errors:  # noqa: WPS329
            raise
        except RunnerError as err:
            logger.error(str(err))
            raise
        except Exception as err:
            logger.error(str(err), exc_info=True)  # noqa: G201
            raise
        finally:
            logger.removeHandler(self.redirect_log_handler)

    @asynccontextmanager
    async def writer(self):
        """Yield a writer.

        Yields:
            A writer.
        """
        read_fd, write_fd = os.pipe()

        reader = asyncio.ensure_future(self._read_from_fd(read_fd))
        writer = FileDescriptorWriter(self, write_fd)

        try:
            yield writer
        finally:
            writer.close()

            with suppress(asyncio.CancelledError):
                await reader

    @asynccontextmanager
    async def redirect_stdout(self):
        """Redirect STDOUT.

        Yields:
            A writer with redirected output.
        """
        async with self.writer() as stdout:
            with redirect_stdout(stdout):
                yield

    @asynccontextmanager
    async def redirect_stderr(self):
        """Redirect STDERR.

        Yields:
            A writer with redirected output.
        """
        async with self.writer() as stderr:
            with redirect_stderr(stderr):
                yield

    def writeline(self, line: str) -> None:
        """Write a line to the underlying structured logger.

        Cleans up any dangling control chars.

        Args:
            line: A line to write.
        """
        self.last_line = line
        self.logger.log(self.write_level, line.rstrip(), name=self.name)

    async def _read_from_fd(self, read_fd):
        # Since we're redirecting our own stdout and stderr output,
        # the line length limit can be arbitrarily large.
        line_length_limit = 1024 * 1024 * 1024  # 1 GiB

        reader = asyncio.StreamReader(limit=line_length_limit)
        read_protocol = asyncio.StreamReaderProtocol(reader)

        loop = asyncio.get_event_loop()
        await loop.connect_read_pipe(lambda: read_protocol, os.fdopen(read_fd))

        await capture_subprocess_output(reader, self)



from __future__ import annotations

import platform

import pytest

from meltano.core.environment import (
    Environment,
    EnvironmentNameContainsStateIdDelimiterError,
)
from meltano.core.environment_service import (
    EnvironmentAlreadyExistsError,
    EnvironmentService,
)
from meltano.core.job_state import STATE_ID_COMPONENT_DELIMITER
from meltano.core.utils import NotFound


class TestEnvironmentService:
    @pytest.fixture()
    def subject(self, environment_service):
        return environment_service

    @pytest.fixture()
    def environment(self, environment_service: EnvironmentService) -> Environment:
        return environment_service.add("test-environment")

    @pytest.mark.order(0)
    def test_add_environment(self, subject: EnvironmentService):
        count = 10
        environments = [Environment(f"environment_{idx}") for idx in range(count)]
        for environment in environments:
            subject.add_environment(environment)

        # default environments are added to initialised projects
        default_environments = [
            Environment(name) for name in ("dev", "staging", "prod")
        ]
        environments = default_environments + environments

        assert subject.list_environments() == environments

        with pytest.raises(
            EnvironmentAlreadyExistsError,
            match="An Environment named 'environment_0' already exists.",
        ):
            subject.add_environment(environments[3])

    @pytest.mark.order(1)
    def test_remove_environment(
        self,
        subject: EnvironmentService,
        environment: Environment,
    ):
        assert subject.list_environments() == [environment]
        subject.remove(environment.name)
        assert not subject.list_environments()

        with pytest.raises(
            NotFound,
            match="Environment 'i-do-not-exist' was not found",
        ):
            subject.remove("i-do-not-exist")

    @pytest.mark.order(2)
    def test_list_environments(
        self,
        subject: EnvironmentService,
    ):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )
        new_environment = subject.add("new-environment")
        assert subject.list_environments() == [new_environment]

    def test_add_name_contains_state_id_component_delimiter(
        self,
        subject: EnvironmentService,
    ):
        with pytest.raises(EnvironmentNameContainsStateIdDelimiterError):
            subject.add(f"test{STATE_ID_COMPONENT_DELIMITER}")



from __future__ import annotations

import fnmatch
import re
import sys
import typing as t
from collections import OrderedDict
from enum import Enum, auto
from functools import singledispatch

import structlog

from meltano.core.behavior.visitor import visit_with

if sys.version_info < (3, 11):
    ReprEnum = Enum
else:
    from enum import ReprEnum

logger = structlog.stdlib.get_logger(__name__)

Node = t.Dict[str, t.Any]
T = t.TypeVar("T", bound="CatalogRule")


UNESCAPED_DOT = re.compile(r"(?<!\\)\.")


class CatalogDict(t.TypedDict):
    """A catalog dictionary."""

    streams: list[dict[str, t.Any]]


class CatalogRule:
    def __init__(
        self,
        tap_stream_id: str | list[str],
        breadcrumb: list[str] | None = None,
        negated: bool = False,
    ):
        """Create a catalog rule for a stream and property."""
        self.tap_stream_id = tap_stream_id
        self.breadcrumb = breadcrumb or []
        self.negated = negated

    @classmethod
    def matching(
        cls: type[T],
        rules: list[T],
        tap_stream_id: str,
        breadcrumb: list[str] | None = None,
    ):
        """Filter rules that match a given breadcrumb."""
        return [rule for rule in rules if rule.match(tap_stream_id, breadcrumb)]

    def match(self, tap_stream_id: str, breadcrumb: list[str] | None = None) -> bool:
        """Evaluate if rule matches a stream or breadcrumb.

        Args:
            tap_stream_id: Singer stream identifier.
            breadcrumb: JSON property breadcrumb.

        Returns:
            Whether the stream ID or breadcrumb matches the rules.
        """
        patterns = (
            self.tap_stream_id
            if isinstance(self.tap_stream_id, list)
            else [self.tap_stream_id]
        )

        result = any(fnmatch.fnmatch(tap_stream_id, pattern) for pattern in patterns)

        # A negated rule matches a stream ID when none of the patterns match
        if self.negated:
            result = not result

        # If provided, the breadcrumb should still match, even on negated rules
        if breadcrumb is not None:
            result = result and fnmatch.fnmatch(
                ".".join(breadcrumb),
                ".".join(self.breadcrumb),
            )

        return result


class MetadataRule(CatalogRule):
    def __init__(
        self,
        tap_stream_id: str | list[str],
        breadcrumb: list[str] | None,
        key: str,
        value: bool,
        negated: bool = False,
    ):
        """Create a metadata rule for a stream and property."""
        super().__init__(tap_stream_id, breadcrumb, negated=negated)
        self.key = key
        self.value = value


class SchemaRule(CatalogRule):
    def __init__(
        self,
        tap_stream_id: str | list[str],
        breadcrumb: list[str] | None,
        payload: dict,
        negated: bool = False,
    ):
        """Create a schema rule for a stream and property."""
        super().__init__(tap_stream_id, breadcrumb, negated=negated)
        self.payload = payload


class SelectPattern(t.NamedTuple):
    """A pattern for selecting streams and properties."""

    stream_pattern: str
    property_pattern: str | None
    negated: bool
    raw: str

    @classmethod
    def parse(cls, pattern: str):
        """Parse a SelectPattern instance from a string pattern.

        Args:
            pattern: Stream or property selection pattern.

        Returns:
            An appropriate `SelectPattern` instance.

        Example:

        >>> SelectPattern.parse("!a.b.c")
        SelectedPattern(
            stream_pattern='a',
            property_pattern='b.c',
            negated=True,
            raw='!a.b.c'
        )
        """
        raw = pattern

        negated = False
        if pattern.startswith("!"):
            negated = True
            pattern = pattern[1:]

        if re.search(UNESCAPED_DOT, pattern):
            stream, prop = re.split(UNESCAPED_DOT, pattern, maxsplit=1)
        else:
            stream = pattern
            prop = None

        return cls(
            stream_pattern=stream.replace(r"\.", "."),
            property_pattern=prop,
            negated=negated,
            raw=raw,
        )


def select_metadata_rules(patterns: t.Iterable[str]) -> list[MetadataRule]:
    """Create metadata rules from `select` patterns.

    Args:
        patterns: Iterable of `select` string patterns.

    Returns:
        A list of corresponding metadata rule objects.
    """
    include_rules = []
    exclude_rules = []

    for pattern in patterns:
        pattern = SelectPattern.parse(pattern)

        prop_pattern = pattern.property_pattern
        selected = not pattern.negated

        rules = include_rules if selected else exclude_rules

        if selected or not prop_pattern or prop_pattern == "*":
            rules.append(
                MetadataRule(
                    tap_stream_id=pattern.stream_pattern,
                    breadcrumb=[],
                    key="selected",
                    value=selected,
                ),
            )

        if prop_pattern:
            props = prop_pattern.split(".")

            rules.append(
                MetadataRule(
                    tap_stream_id=pattern.stream_pattern,
                    breadcrumb=property_breadcrumb(props),
                    key="selected",
                    value=selected,
                ),
            )

    return include_rules + exclude_rules


def select_filter_metadata_rules(patterns: t.Iterable[str]) -> list[MetadataRule]:
    """Create metadata rules from `select_filter` patterns.

    Args:
        patterns: Iterable of `select_filter` string patterns.

    Returns:
        A list of corresponding metadata rule objects.
    """
    # We set `selected: false` if the `tap_stream_id`
    # does NOT match any of the selection/inclusion patterns
    include_rule = MetadataRule(
        negated=True,
        tap_stream_id=[],
        breadcrumb=[],
        key="selected",
        value=False,
    )
    # Or if it matches one of the exclusion patterns
    exclude_rule = MetadataRule(
        tap_stream_id=[],
        breadcrumb=[],
        key="selected",
        value=False,
    )

    for pattern in patterns:
        pattern = SelectPattern.parse(pattern)

        rule = exclude_rule if pattern.negated else include_rule
        rule.tap_stream_id.append(pattern.stream_pattern)

    rules = []
    if include_rule.tap_stream_id:
        rules.append(include_rule)
    if exclude_rule.tap_stream_id:
        rules.append(exclude_rule)

    return rules


def path_property(path: str) -> str:
    """Extract the property name from a materialized path.

    As we traverse the catalog tree, we build a materialized path
    to keep track of the parent nodes.

    Args:
        path: String representing a property path in the JSON schema.

    Returns:
        A string representing a property path in the JSON object.

    Examples:
      stream[0].properties.list_items.properties.account → list_items.account
      stream[0].properties.name                          → name
    """
    prop_regex = r"properties\.([^.]+)+"
    components = re.findall(prop_regex, path)
    return ".".join(components)


def property_breadcrumb(props: list[str]) -> list[str]:
    """Create breadcrumb from properties path list.

    Args:
        props: List of strings representing a property breadcrumb in the JSON object.

    Returns:
        A list of strings representing a property breadcrumb in the JSON schema.

    Example:
    >>> property_breadcrumb(["payload", "content"])
    ['properties', 'payload', 'properties', 'content']
    """
    if len(props) >= 2 and props[0] == "properties":
        breadcrumb = props
    else:
        breadcrumb = []
        for prop in props:
            breadcrumb.extend(["properties", prop])

    return breadcrumb


class CatalogNode(Enum):
    STREAM = auto()
    PROPERTY = auto()
    METADATA = auto()


# TODO: Move to `enum.StrEnum` when support for Python 3.8 is dropped
class SelectionType(str, ReprEnum):
    """A valid stream or property selection type."""

    SELECTED = "selected"
    EXCLUDED = "excluded"
    AUTOMATIC = "automatic"

    def __bool__(self):
        return self is not self.__class__.EXCLUDED

    def __add__(self, other):
        if self is SelectionType.EXCLUDED or other is SelectionType.EXCLUDED:
            return SelectionType.EXCLUDED

        if self is SelectionType.AUTOMATIC or other is SelectionType.AUTOMATIC:
            return SelectionType.AUTOMATIC

        return SelectionType.SELECTED


@singledispatch
def visit(  # noqa: D103
    node,  # noqa: ARG001
    executor,  # noqa: ARG001
    path: str = "",
):
    logger.debug("Skipping node at '%s'", path)  # noqa: WPS323


@visit.register(dict)
def _(node: dict, executor, path=""):
    node_type = None

    if re.search(r"streams\[\d+\]$", path):
        node_type = CatalogNode.STREAM

    if re.search(r"schema(\.properties\.\w*)+$", path):
        node_type = CatalogNode.PROPERTY

    if re.search(r"metadata\[\d+\]$", path) and "breadcrumb" in node:
        node_type = CatalogNode.METADATA

    if node_type:
        logger.debug("Visiting %s at '%s'.", node_type, path)  # noqa: WPS323
        executor(node_type, node, path)

    for child_path, child_node in node.items():
        if node_type is CatalogNode.PROPERTY and child_path in {"anyOf", "type"}:
            continue

        # TODO mbergeron: refactor this to use a dynamic visitor per CatalogNode
        executor.visit(child_node, path=f"{path}.{child_path}")


@visit.register(list)
def _(node: list, executor, path=""):
    for index, child_node in enumerate(node):
        executor.visit(child_node, path=f"{path}[{index}]")


@visit_with(visit)
class CatalogExecutor:
    def execute(self, node_type: CatalogNode, node: Node, path: str):
        """Dispatch all node methods."""
        dispatch = {
            CatalogNode.STREAM: self.stream_node,
            CatalogNode.PROPERTY: self.property_node,
            CatalogNode.METADATA: self.metadata_node,
        }

        try:
            dispatch[node_type](node, path)
        except KeyError:  # pragma: no cover
            logger.debug("Unknown node type '%s'.", node_type)  # noqa: WPS323

    def stream_node(self, node: Node, path: str):
        """Process stream node."""

    def property_node(self, node: Node, path: str):
        """Process property node."""

    def metadata_node(self, node: Node, path: str):
        """Process metadata node."""
        if len(node["breadcrumb"]) == 0:
            self.stream_metadata_node(node, path)
        else:
            self.property_metadata_node(node, path)

    def stream_metadata_node(self, node: Node, path: str):
        """Process stream metadata node."""

    def property_metadata_node(self, node: Node, path: str):
        """Process property metadata node."""

    def __call__(self, node_type, node: Node, path: str):
        """Call this instance as a function."""
        return self.execute(node_type, node, path)


class MetadataExecutor(CatalogExecutor):
    def __init__(self, rules: list[MetadataRule]):
        self._stream = None
        self._rules = rules

    def ensure_metadata(self, breadcrumb: list[str]):
        """Handle missing metadata entries."""
        metadata_list: list[dict] = self._stream["metadata"]
        match = next(
            (
                metadata
                for metadata in metadata_list
                if metadata["breadcrumb"] == breadcrumb
            ),
            None,
        )

        # Missing inclusion metadata for property
        if match is None:
            # Streams and top-level properties.
            if len(breadcrumb) <= 2:
                entry = {
                    "breadcrumb": breadcrumb,
                    "metadata": {"inclusion": "automatic"},
                }
            # Exclude nested properties.
            else:
                entry = {
                    "breadcrumb": breadcrumb,
                    "metadata": {"inclusion": "available"},
                }

            metadata_list.append(entry)

    def stream_node(self, node: Node, path: str):
        """Process stream metadata node."""
        self._stream = node
        tap_stream_id = self._stream["tap_stream_id"]

        if "metadata" not in node:
            node["metadata"] = []

        self.ensure_metadata([])

        for rule in MetadataRule.matching(self._rules, tap_stream_id, []):
            # Legacy catalogs have underscorized keys on the streams themselves
            self.set_metadata(node, path, rule.key.replace("-", "_"), rule.value)

    def property_node(
        self,
        node: Node,  # noqa: ARG002
        path: str,
    ):
        """Process property metadata node."""
        breadcrumb_idx = path.index("properties")
        breadcrumb = path[breadcrumb_idx:].split(".")

        self.ensure_metadata(breadcrumb)

    def metadata_node(self, node: Node, path: str):
        """Process metadata node."""
        tap_stream_id = self._stream["tap_stream_id"]
        breadcrumb = node["breadcrumb"]

        logger.debug(
            "Visiting metadata node for tap_stream_id '%s', breadcrumb '%s'",  # noqa: WPS323, E501
            tap_stream_id,
            breadcrumb,
        )

        for rule in MetadataRule.matching(self._rules, tap_stream_id, breadcrumb):
            self.set_metadata(
                node["metadata"],
                f"{path}.metadata",
                rule.key,
                rule.value,
            )

    def set_metadata(self, node: Node, path: str, key: str, value: t.Any):
        """Set selection and inclusion keys in a metadata node."""
        # Unsupported fields cannot be selected
        if (
            key == "selected"
            and value is True
            and node.get("inclusion") == "unsupported"
        ):
            return

        node[key] = value
        logger.debug("Setting '%s.%s' to '%s'", path, key, value)  # noqa: WPS323


class SelectExecutor(MetadataExecutor):
    def __init__(self, patterns: list[str]):
        super().__init__(select_metadata_rules(patterns))


class SchemaExecutor(CatalogExecutor):
    def __init__(self, rules: list[SchemaRule]):
        self._stream = None
        self._rules = rules

    def ensure_property(self, breadcrumb: list[str]):  # noqa: WPS231
        """Create nodes for the breadcrumb and schema extra that matches."""
        next_node: dict[str, t.Any] = self._stream["schema"]

        for idx, key in enumerate(breadcrumb):
            # If the key contains shell-style wildcards,
            # ensure property nodes exist for matching breadcrumbs.
            if re.match(r"[*?\[\]]", key):
                node_keys = next_node.keys()
                if matching_keys := fnmatch.filter(node_keys, key):
                    matching_breadcrumb = breadcrumb.copy()
                    for key in matching_keys:
                        matching_breadcrumb[idx] = key
                        self.ensure_property(matching_breadcrumb)

                break

            # If a property node for this breadcrumb doesn't exist yet, create it.
            if key not in next_node:
                next_node[key] = {}

            next_node = next_node[key]

    def stream_node(
        self,
        node: Node,
        path,  # noqa: ARG002
    ):
        """Process stream schema node."""
        self._stream = node
        tap_stream_id: str = self._stream["tap_stream_id"]

        if "schema" not in node:
            node["schema"] = {"type": "object"}

        for rule in SchemaRule.matching(self._rules, tap_stream_id):
            self.ensure_property(rule.breadcrumb)

    def property_node(self, node: Node, path: str):
        """Process property schema node."""
        tap_stream_id = self._stream["tap_stream_id"]

        breadcrumb_idx = path.index("properties")
        breadcrumb = path[breadcrumb_idx:].split(".")

        for rule in SchemaRule.matching(self._rules, tap_stream_id, breadcrumb):
            self.set_payload(node, path, rule.payload)

    def set_payload(self, node: Node, path: str, payload: dict):
        """Set node payload from a clean mapping."""
        node.clear()
        node.update(payload)
        logger.debug("Setting '%s' to %r", path, payload)  # noqa: WPS323


class ListExecutor(CatalogExecutor):
    def __init__(self):
        # properties per stream
        self.properties: dict[str, set[str]] = OrderedDict()

        super().__init__()

    def stream_node(
        self,
        node: Node,
        path: str,  # noqa: ARG002
    ):
        """Initialize empty property set stream."""
        stream = node["tap_stream_id"]
        if stream not in self.properties:
            self.properties[stream] = set()

    def property_node(
        self,
        node: Node,  # noqa: ARG002
        path: str,
    ):
        """Add property to stream collection."""
        prop = path_property(path)
        # current stream
        stream = next(reversed(self.properties))
        self.properties[stream].add(prop)


class SelectedNode(t.NamedTuple):
    """Selection type and key of a node."""

    key: str
    selection: SelectionType


class ListSelectedExecutor(CatalogExecutor):
    def __init__(self):
        self.streams: set[SelectedNode] = set()
        self.properties: dict[str, set[SelectedNode]] = OrderedDict()
        super().__init__()

    @property
    def selected_properties(self):
        """Get selected streams and properties."""
        # we don't want to mutate the visitor result
        selected_properties = self.properties.copy()

        # remove all non-selected streams
        for name, selected in self.streams:
            if not selected:
                del selected_properties[name]

        # remove all non-selected properties
        for stream, props in selected_properties.items():
            selected_properties[stream] = {name for name, selected in props if selected}

        return selected_properties

    @staticmethod
    def node_selection(node: Node) -> SelectionType:
        """Get selection type from metadata entry.

        Args:
            node: Catalog metadata dictionary.

        Returns:
            A proper `SelectionType` given the inclusion and selection metadata.
        """
        try:
            metadata: dict[str, t.Any] = node["metadata"]
        except KeyError:
            return SelectionType.EXCLUDED

        if metadata.get("inclusion") == "automatic":
            return SelectionType.AUTOMATIC
        if metadata.get("selected") is True or (
            metadata.get("selected") is None
            and metadata.get("selected-by-default", False)
        ):
            return SelectionType.SELECTED
        return SelectionType.EXCLUDED

    def stream_node(
        self,
        node: Node,
        path: str,  # noqa: ARG002
    ):
        """Initialize empty set for selected nodes in stream."""
        self._stream: str = node["tap_stream_id"]
        self.properties[self._stream] = set()

    def stream_metadata_node(
        self,
        node: Node,
        path: str,  # noqa: ARG002
    ):
        """Add stream selection to tap's collection."""
        selection = SelectedNode(self._stream, self.node_selection(node))
        self.streams.add(selection)

    def property_metadata_node(
        self,
        node: Node,
        path: str,  # noqa: ARG002
    ):
        """Add property selection to stream's collection."""
        property_path = ".".join(node["breadcrumb"])
        prop = path_property(property_path)
        selection = SelectedNode(prop, self.node_selection(node))

        self.properties[self._stream].add(selection)



"""Defines Job model class."""

from __future__ import annotations

import asyncio
import os
import signal
import typing as t
import uuid
from contextlib import asynccontextmanager, contextmanager, suppress
from datetime import datetime, timedelta, timezone
from enum import Enum
from enum import IntFlag as EnumIntFlag

from sqlalchemy import literal
from sqlalchemy.ext.hybrid import Comparator, hybrid_property
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import Mapped, mapped_column

from meltano.core.error import Error
from meltano.core.models import SystemModel
from meltano.core.sqlalchemy import (
    DateTimeUTC,
    GUIDType,
    IntFlag,
    IntPK,
    JSONEncodedDict,
)

HEARTBEATLESS_JOB_VALID_HOURS = 24
HEARTBEAT_VALID_MINUTES = 5


class InconsistentStateError(Error):
    """Occur upon a wrong operation for the current state."""


class ImpossibleTransitionError(Error):
    """Occur upon a wrong transition."""


class State(Enum):
    """Represents status of a Job."""

    IDLE = (0, ("RUNNING", "FAIL"))
    RUNNING = (1, ("SUCCESS", "FAIL"))
    SUCCESS = (2, ())
    FAIL = (3, ("RUNNING",))
    DEAD = (4, ())
    STATE_EDIT = (5, ())

    def transitions(self):
        """Get possible next States for a job of this State.

        Returns:
            The possible states jobs in this state can be transitioned into
        """
        return self.value[1]

    def __str__(self):
        """Get a string representation of this State.

        Returns:
            the name of this State
        """
        return self.name


class StateComparator(Comparator):
    """Compare Job._state to State enums."""

    def __eq__(self, other):
        """Enable SQLAlchemy to directly compare Job.state values with State.

        Args:
            other: the State enum to compare to

        Returns:
            Comparison between __clause_element__ and SQLAlchemy literal for State name
        """
        return self.__clause_element__() == literal(other.name)


def current_trigger():
    """Get the trigger for running job.

    Returns:
        The trigger for currently running job
    """
    return os.getenv("MELTANO_JOB_TRIGGER")


class Payload(EnumIntFlag):
    """Flag indicating whether a Job has state in its payload field."""

    STATE = 1
    INCOMPLETE_STATE = 2


class Job(SystemModel):  # noqa: WPS214
    """Model class that represents a `meltano elt` run in the system database.

    Includes State.STATE_EDIT rows which represent CLI invocations of the
    `meltano state` command which wrote state to the db. Queries that are
    meant to return only actual job runs should filter out records with
    state == State.STATE_EDIT.
    """

    __tablename__ = "runs"

    id: Mapped[IntPK]
    job_name: Mapped[t.Optional[str]]  # noqa: UP007
    run_id: Mapped[GUIDType]
    _state: Mapped[t.Optional[str]] = mapped_column(name="state")  # noqa: UP007
    started_at: Mapped[t.Optional[datetime]] = mapped_column(DateTimeUTC)  # noqa: UP007
    last_heartbeat_at: Mapped[t.Optional[datetime]] = mapped_column(  # noqa: UP007
        DateTimeUTC,
    )
    ended_at: Mapped[t.Optional[datetime]] = mapped_column(DateTimeUTC)  # noqa: UP007
    payload: Mapped[dict] = mapped_column(MutableDict.as_mutable(JSONEncodedDict))
    payload_flags: Mapped[Payload] = mapped_column(IntFlag, default=0)
    trigger: Mapped[t.Optional[str]] = mapped_column(  # noqa: UP007
        default=current_trigger,
    )

    def __init__(self, **kwargs):
        """Construct a Job.

        Args:
            kwargs: keyword args to override defaults and pass to super
        """
        kwargs["_state"] = kwargs.pop("state", State.IDLE).name
        kwargs["payload"] = kwargs.get("payload", {})
        kwargs["run_id"] = kwargs.get("run_id", uuid.uuid4())
        super().__init__(**kwargs)

    @hybrid_property
    def state(self) -> State:
        """Get the job state as a State enum.

        Returns:
            State enum matching string value for this job state
        """
        return State[self._state]

    @state.setter
    def state(self, value):
        """Set the _state value for this Job from a State enum.

        Args:
            value: the State enum to use.
        """
        self._state = str(value)

    @state.comparator
    def state(cls):  # noqa: N805
        """Use this comparison to compare Job.state to State.

        See:
            https://docs.sqlalchemy.org/en/14/orm/extensions/hybrid.html#building-custom-comparators

        Returns:
            Result of comparison
        """
        return StateComparator(cls._state)

    def is_running(self):
        """Return whether Job is running.

        Returns:
            bool indicating whether this Job is running
        """
        return self.state is State.RUNNING

    def is_stale(self):
        """Return whether Job has gone stale.

        Running jobs with a heartbeat are considered stale after no heartbeat
        is recorded for 5 minutes.
        Legacy jobs without a heartbeat are considered stale after being in
        the running state for 24 hours.

        Returns:
            bool indicating whether this Job is stale
        """
        if not self.is_running():
            return False

        if self.last_heartbeat_at:
            timestamp = self.last_heartbeat_at
            valid_for = timedelta(minutes=HEARTBEAT_VALID_MINUTES)
        else:
            timestamp = self.started_at
            valid_for = timedelta(hours=HEARTBEATLESS_JOB_VALID_HOURS)

        return datetime.now(timezone.utc) - timestamp > valid_for

    def has_error(self):
        """Return whether a job has failed.

        Returns:
            bool indicating whether this Job has failed
        """
        return self.state is State.FAIL

    def is_complete(self):
        """Return whether a job has completed.

        Returns:
            bool indicating whether this job has completed
        """
        return self.state in {State.SUCCESS, State.FAIL}

    def is_success(self):
        """Return whether a job has succeeded.

        Returns:
            a bool indicating whether this job has succeeded
        """
        return self.state is State.SUCCESS

    def can_transit(self, state: State) -> bool:
        """Return whether this job can transit into the given state.

        Args:
            state: the state to check against

        Returns:
            bool indicating whether the given state is transitable from this job's state
        """
        return True if self.state is state else state.name in self.state.transitions()

    def transit(self, state: State) -> tuple[State, State]:
        """Transition this job into the given state.

        Args:
            state: The state to transition this job to

        Returns:
            A tuple with the original state and the new state

        Raises:
            ImpossibleTransitionError: when this job cannot transition into the
                given state
        """
        transition = (self.state, state)

        if not self.can_transit(state):
            raise ImpossibleTransitionError(transition)

        if self.state is state:
            return transition

        self.state = state

        return transition

    @asynccontextmanager
    async def run(self, session):
        """Run wrapped code in context of a job.

        Transitions state to RUNNING and SUCCESS/FAIL as appropriate and
        records heartbeat every second.

        Args:
            session: the session to use for writing to the db

        Raises:
            BaseException: re-raises an exception occurring in the job running
                in this context
        """  # noqa: DAR301
        try:
            self.start()
            self.save(session)

            with self._handling_sigterm(session):
                async with self._heartbeating(session):
                    yield

            self.success()
            self.save(session)
        except BaseException as err:  # noqa: WPS424
            if not self.is_running():
                raise

            self.fail(error=self._error_message(err))
            self.save(session)
            raise

    def start(self):
        """Mark the job has having started."""
        self.started_at = datetime.now(timezone.utc)
        self.transit(State.RUNNING)

    def fail(self, error=None):
        """Mark the job as having failed.

        Args:
            error: the error to associate with the job's failure
        """
        self.ended_at = datetime.now(timezone.utc)
        self.transit(State.FAIL)
        if error:
            self.payload.update({"error": str(error)})

    def success(self):
        """Mark the job as having succeeded."""
        self.ended_at = datetime.now(timezone.utc)
        self.transit(State.SUCCESS)

    def fail_stale(self):
        """Mark job as failed if it's gone stale.

        Returns:
            False if job is not stale, else True
        """
        if not self.is_stale():
            return False

        if self.last_heartbeat_at:
            reason = f"No heartbeat recorded for {HEARTBEAT_VALID_MINUTES} minutes."
        else:
            reason = f"Still running after {HEARTBEATLESS_JOB_VALID_HOURS} hours."

        self.fail(f"{reason} The process was likely killed unceremoniously.")

        return True

    def __repr__(self):
        """Represent as a string.

        Returns:
            A string representation of the job.
        """
        return (
            f"<Job(id='{self.id}', job_name='{self.job_name}', "
            f"state='{self.state}', started_at='{self.started_at}', "
            f"ended_at='{self.ended_at}')>"
        )

    def save(self, session):
        """Save the job in the db.

        Args:
            session: the session to use in querying the db

        Returns:
            the saved job
        """
        session.add(self)
        session.commit()

        return self

    def _heartbeat(self):
        """Update last_heartbeat_at for this job in the db."""
        self.last_heartbeat_at = datetime.now(timezone.utc)

    async def _heartbeater(self, session):
        """Heartbeat to the db every second.

        Args:
            session: the session to use for writing to the db
        """
        while True:  # noqa: WPS457
            self._heartbeat()
            self.save(session)

            await asyncio.sleep(1)

    @asynccontextmanager
    async def _heartbeating(self, session):
        """Provide a context for heartbeating jobs.

        Args:
            session: the session to use for writing to the db
        """  # noqa: DAR301
        heartbeat_future = asyncio.ensure_future(self._heartbeater(session))
        try:
            yield
        finally:
            if not heartbeat_future.cancelled():
                heartbeat_future.cancel()

            with suppress(asyncio.CancelledError):
                await heartbeat_future

    @contextmanager
    def _handling_sigterm(
        self,
        session,  # noqa: ARG002
    ):
        def handler(*_):  # noqa: WPS430
            sigterm_status = 143
            raise SystemExit(sigterm_status)

        original_termination_handler = signal.signal(signal.SIGTERM, handler)

        try:
            yield
        finally:
            signal.signal(signal.SIGTERM, original_termination_handler)

    def _error_message(self, err):
        if isinstance(err, SystemExit):
            return "The process was terminated"

        if isinstance(err, (KeyboardInterrupt, asyncio.CancelledError)):
            return "The process was interrupted"

        if str(err):
            return str(err)

        return repr(err)



from __future__ import annotations

from datetime import datetime, timedelta, timezone

import pytest

from meltano.core.job.finder import JobFinder
from meltano.core.job.job import (
    HEARTBEAT_VALID_MINUTES,
    HEARTBEATLESS_JOB_VALID_HOURS,
    Job,
    State,
)


class TestJobFinder:
    @pytest.mark.parametrize(
        ("state", "started_at_hours_ago", "last_heartbeat_at_minutes_ago", "is_stale"),
        (
            (State.RUNNING, None, None, False),
            (State.RUNNING, HEARTBEATLESS_JOB_VALID_HOURS - 1, None, False),
            (State.RUNNING, HEARTBEATLESS_JOB_VALID_HOURS + 1, None, True),
            (State.RUNNING, None, 0, False),
            (State.RUNNING, None, HEARTBEAT_VALID_MINUTES - 1, False),
            (State.RUNNING, None, HEARTBEAT_VALID_MINUTES + 1, True),
            (State.SUCCESS, None, None, False),
            (State.FAIL, None, None, False),
        ),
    )
    def test_all_stale(
        self,
        state,
        started_at_hours_ago,
        last_heartbeat_at_minutes_ago,
        is_stale,
        session,
    ):
        started_at = datetime.now(timezone.utc)
        if started_at_hours_ago:
            started_at -= timedelta(hours=started_at_hours_ago)

        last_heartbeat_at = None
        if last_heartbeat_at_minutes_ago is not None:
            last_heartbeat_at = datetime.now(timezone.utc) - timedelta(
                minutes=last_heartbeat_at_minutes_ago,
            )

        job = Job(
            job_name="test_state_id",
            state=state,
            started_at=started_at,
            last_heartbeat_at=last_heartbeat_at,
        )
        job.save(session)

        assert job.is_stale() == is_stale

        assert bool(job in JobFinder.all_stale(session)) == is_stale

    def test_stale(self, session):
        job = Job(job_name="test")
        job.start()
        job.last_heartbeat_at = datetime.now(timezone.utc) - timedelta(minutes=10)
        job.save(session)

        assert job in JobFinder(state_id=job.job_name).stale(session)

        assert job not in JobFinder(state_id="other").stale(session)



"""StateStoreManager for Google Cloud storage backend."""

from __future__ import annotations

import typing as t
from contextlib import contextmanager
from functools import cached_property

from meltano.core.state_store.filesystem import CloudStateStoreManager

if t.TYPE_CHECKING:
    from collections.abc import Iterator

GOOGLE_INSTALLED = True

try:
    import google
    import google.api_core.exceptions  # noqa: WPS301
    import google.cloud.storage  # type: ignore[import-untyped]  # noqa: WPS301
except ImportError:
    GOOGLE_INSTALLED = False


class MissingGoogleError(Exception):
    """Raised when google is required but not installed."""

    def __init__(self):
        """Initialize a MissingGoogleError."""
        super().__init__(
            "google-cloud-storage required but not installed. Install meltano[gcs] to use GCS as a state backend.",  # noqa: E501
        )


@contextmanager
def requires_gcs():
    """Raise MissingGoogleError if gcs is required but missing in context.

    Raises:
        MissingGoogleError: if google-cloud-storage is not installed.

    Yields:
        None
    """
    if not GOOGLE_INSTALLED:
        raise MissingGoogleError
    yield


class GCSStateStoreManager(CloudStateStoreManager):
    """State backend for Google Cloud Storage."""

    label: str = "Google Cloud Storage"

    def __init__(
        self,
        bucket: str | None = None,
        prefix: str | None = None,
        application_credentials: str | None = None,
        **kwargs,
    ):
        """Initialize the BaseFilesystemStateStoreManager.

        Args:
            bucket: the bucket to store state in
            prefix: the prefix to store state at
            application_credentials: application credentials to use in
                authenticating to GCS
            kwargs: additional keyword args to pass to parent
        """
        super().__init__(**kwargs)
        self.bucket = bucket or self.parsed.hostname
        self.prefix = prefix or self.parsed.path
        self.application_credentials = application_credentials

    @staticmethod
    @requires_gcs()
    def is_file_not_found_error(err: Exception) -> bool:
        """Check if err is equivalent to file not being found.

        Args:
            err: the err to check

        Returns:
            True if error represents file not being found, else False
        """
        return isinstance(err, google.api_core.exceptions.NotFound) and (
            "No such object:" in err.args[0] or "blob" in err.args[0]
        )

    @cached_property
    def client(self):
        """Get an authenticated google.cloud.storage.Client.

        Returns:
            A google.cloud.storage.Client.
        """
        with requires_gcs():
            if self.application_credentials:
                return google.cloud.storage.Client.from_service_account_json(
                    self.application_credentials,
                )
            # Use default authentication in environment
            return google.cloud.storage.Client()

    def delete(self, file_path: str):
        """Delete the file/blob at the given path.

        Args:
            file_path: the path to delete.

        Raises:
            Exception: if error not indicating file is not found is thrown
        """
        bucket = self.client.bucket(self.bucket)
        try:
            blob = bucket.blob(file_path)
            blob.delete()
        except Exception as e:
            if self.is_file_not_found_error(e):
                ...
            else:
                raise e

    def list_all_files(self) -> Iterator[str]:
        """List all files in the backend.

        Yields:
            The next file in the backend.
        """
        for blob in self.client.list_blobs(  # noqa: WPS526
            bucket_or_name=self.bucket, prefix=self.state_dir
        ):
            yield blob.name

    def copy_file(self, src: str, dst: str) -> None:
        """Copy a file from one location to another.

        Args:
            src: the source path
            dst: the destination path
        """
        bucket = self.client.bucket(self.bucket)
        blob = bucket.blob(src)
        bucket.copy_blob(blob, bucket, dst)
